#
# Copyright 2018-2023 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json

import tornado


def expected_http_error(error, expected_code, expected_message=None):
    """Check that the error matches the expected output error."""
    e = error.value
    if isinstance(e, tornado.web.HTTPError):
        if expected_code != e.status_code:
            return False
        if expected_message is not None and expected_message != str(e):
            return False
        return True
    elif any(
        [
            isinstance(e, tornado.httpclient.HTTPClientError),
            isinstance(e, tornado.httpclient.HTTPError),
        ]
    ):
        if expected_code != e.code:
            return False
        if expected_message:
            message = json.loads(e.response.body.decode())["message"]
            if expected_message != message:
                return False
        return True



#
# Copyright 2018-2023 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import shutil

import pytest


@pytest.fixture
def setup_factory_data(jp_environ, jp_env_jupyter_path):
    """Copies the factory metadata instances for runtime-images and compontent-registries to test hierarchy."""
    source = os.path.join(os.path.dirname(__file__), "..", "..", "..", "etc/config/metadata")
    destination = os.path.join(jp_env_jupyter_path, "metadata")
    shutil.copytree(source, destination)
    yield destination  # this return value probably won't be used, but here nonetheless


# Set Elyra server extension as enabled (overriding server_config fixture from jupyter_server)
@pytest.fixture
def jp_server_config(setup_factory_data):
    return {"ServerApp": {"jpserver_extensions": {"elyra": True}}}



#
# Copyright 2018-2023 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

PIPELINE_DEFAULTS = "pipeline_defaults"
PIPELINE_PARAMETERS = "pipeline_parameters"
RUNTIME_IMAGE = "runtime_image"
ENV_VARIABLES = "env_vars"
MOUNTED_VOLUMES = "mounted_volumes"
KUBERNETES_SECRETS = "kubernetes_secrets"
KUBERNETES_TOLERATIONS = "kubernetes_tolerations"
KUBERNETES_POD_ANNOTATIONS = "kubernetes_pod_annotations"
KUBERNETES_POD_LABELS = "kubernetes_pod_labels"
DISABLE_NODE_CACHING = "disable_node_caching"
KUBERNETES_SHARED_MEM_SIZE = "kubernetes_shared_mem_size"
COS_OBJECT_PREFIX = "cos_object_prefix"  # optional static prefix to be used when generating object name for cos storage



#
# Copyright 2018-2023 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import os
from pathlib import Path
from subprocess import CompletedProcess
from subprocess import run

from conftest import KFP_COMPONENT_CACHE_INSTANCE
from conftest import TEST_CATALOG_NAME
import jupyter_core.paths
import pytest
import yaml

from elyra.metadata.metadata import Metadata
from elyra.pipeline.catalog_connector import CatalogEntry
from elyra.pipeline.catalog_connector import EntryData
from elyra.pipeline.catalog_connector import FilesystemComponentCatalogConnector
from elyra.pipeline.catalog_connector import UrlComponentCatalogConnector
from elyra.pipeline.component_catalog import ComponentCache
from elyra.pipeline.component_metadata import ComponentCatalogMetadata
from elyra.pipeline.kfp.component_parser_kfp import KfpComponentParser
from elyra.pipeline.runtime_type import RuntimeProcessorType

COMPONENT_CATALOG_DIRECTORY = os.path.join(jupyter_core.paths.ENV_JUPYTER_PATH[0], "components")
RUNTIME_PROCESSOR = RuntimeProcessorType.KUBEFLOW_PIPELINES


def _get_resource_path(filename):
    pipeline_dir = os.path.realpath(os.path.dirname(os.path.dirname(__file__)))
    resource_path = os.path.join(pipeline_dir, "resources", "components", filename)
    resource_path = os.path.normpath(resource_path)

    return resource_path


@pytest.mark.parametrize("catalog_instance", [KFP_COMPONENT_CACHE_INSTANCE], indirect=True)
def test_component_catalog_load(component_cache, catalog_instance):
    components = component_cache.get_all_components(RUNTIME_PROCESSOR)
    assert len(components) > 0


@pytest.mark.parametrize("create_inprocess", [True, False])
async def test_modify_component_catalogs(jp_environ, component_cache, metadata_manager_with_teardown, create_inprocess):
    # Get initial set of components
    initial_components = component_cache.get_all_components(RUNTIME_PROCESSOR)

    # Create new registry instance with a single URL-based component
    paths = [_get_resource_path("kfp_test_operator.yaml")]

    instance_metadata = {
        "description": "A test registry",
        "runtime_type": RUNTIME_PROCESSOR.name,
        "categories": ["New Components"],
        "paths": paths,
    }
    registry_instance = Metadata(
        schema_name="local-file-catalog",
        name=TEST_CATALOG_NAME,
        display_name="New Test Registry",
        metadata=instance_metadata,
    )

    if create_inprocess:
        metadata_manager_with_teardown.create(TEST_CATALOG_NAME, registry_instance)
    else:
        res: CompletedProcess = run(
            [
                "elyra-metadata",
                "create",
                "component-catalogs",
                f"--schema_name={registry_instance.schema_name}",
                f"--json={registry_instance.to_json()}",
                f"--name={TEST_CATALOG_NAME}",
            ]
        )
        assert res.returncode == 0

    # Wait for update to complete
    component_cache.wait_for_all_cache_tasks()

    # Get new set of components from all active registries, including added test registry
    components_after_create = component_cache.get_all_components(RUNTIME_PROCESSOR)
    assert len(components_after_create) == len(initial_components) + 1

    added_component_names = [component.name for component in components_after_create]
    assert "Test Operator" in added_component_names
    assert "Test Operator No Inputs" not in added_component_names

    # Modify the test registry to add a path to the catalog instance
    paths.append(_get_resource_path("kfp_test_operator_no_inputs.yaml"))
    metadata_manager_with_teardown.update(TEST_CATALOG_NAME, registry_instance)

    # Wait for update to complete
    component_cache.wait_for_all_cache_tasks()

    # Get set of components from all active registries, including modified test registry
    components_after_update = component_cache.get_all_components(RUNTIME_PROCESSOR)
    assert len(components_after_update) == len(initial_components) + 2

    modified_component_names = [component.name for component in components_after_update]
    assert "Test Operator" in modified_component_names
    assert "Test Operator No Inputs" in modified_component_names

    # Delete the test registry
    metadata_manager_with_teardown.remove(TEST_CATALOG_NAME)

    # Wait for update to complete
    component_cache.wait_for_all_cache_tasks()

    # Check that components remaining after delete are the same as before the new catalog was added
    components_after_remove = component_cache.get_all_components(RUNTIME_PROCESSOR)
    assert len(components_after_remove) == len(initial_components)


@pytest.mark.parametrize("create_inprocess", [True, False])
async def test_directory_based_component_catalog(
    component_cache, metadata_manager_with_teardown, create_inprocess, tmpdir
):
    # Verify that the component cache is empty to prevent other tests
    # from having an impact on this' tests result
    initial_components = component_cache.get_all_components(RUNTIME_PROCESSOR)
    assert len(initial_components) == 0, initial_components[0].name

    # Create and populate a temporary catalog directory
    catalog_dir = Path(tmpdir) / "catalog"
    catalog_dir.mkdir()
    # Copy a few YAML files from ../resources/components to
    # the catalog directory
    directory_entries = {"download_data.yaml": None, "kfp_test_operator_no_inputs.yaml": None}
    for file in directory_entries:
        with open(_get_resource_path(file), "r") as fh_in:
            # read file
            data = fh_in.read()
            # extract and store component name
            directory_entries[file] = yaml.safe_load(data)["name"]
            # write (unchanged) file to destination
            with open(catalog_dir / file, "w") as fh_out:
                fh_out.write(data)
        # make sure the file exists in the destination
        assert (catalog_dir / file).is_file()

    # Create new directory-based registry
    instance_metadata = {
        "description": "A test registry",
        "runtime_type": RUNTIME_PROCESSOR.name,
        "categories": ["New Components"],
        "paths": [str(catalog_dir)],
    }
    registry_instance = Metadata(
        schema_name="local-directory-catalog",
        name=TEST_CATALOG_NAME,
        display_name="New Test Registry",
        metadata=instance_metadata,
    )

    if create_inprocess:
        metadata_manager_with_teardown.create(TEST_CATALOG_NAME, registry_instance)
    else:
        res: CompletedProcess = run(
            [
                "elyra-metadata",
                "create",
                "component-catalogs",
                f"--schema_name={registry_instance.schema_name}",
                f"--json={registry_instance.to_json()}",
                f"--name={TEST_CATALOG_NAME}",
            ]
        )
        assert res.returncode == 0

    # Wait for update to complete
    component_cache.wait_for_all_cache_tasks()

    # Verify that the number of components in the cache equals the number of
    # components in the directory catalog
    components_after_create = component_cache.get_all_components(RUNTIME_PROCESSOR)
    assert len(components_after_create) == len(directory_entries), components_after_create

    # Verify the component names
    added_component_names = [component.name for component in components_after_create]
    for component in directory_entries:
        assert directory_entries[component] in added_component_names

    # Delete the test registry and wait for updates to complete
    metadata_manager_with_teardown.remove(TEST_CATALOG_NAME)
    component_cache.wait_for_all_cache_tasks()


def test_parse_kfp_component_file():
    # Define the appropriate reader for a filesystem-type component definition
    kfp_supported_file_types = [".yaml"]
    reader = FilesystemComponentCatalogConnector(kfp_supported_file_types)

    # Read contents of given path
    path = _get_resource_path("kfp_test_operator.yaml")
    catalog_entry_data = {"path": path}

    # Construct a catalog instance
    catalog_type = "local-file-catalog"
    catalog_instance = ComponentCatalogMetadata(
        schema_name=catalog_type, metadata={"categories": ["Test"], "runtime_type": RUNTIME_PROCESSOR.name}
    )

    # Build the catalog entry data structures required for parsing
    entry_data = reader.get_entry_data(catalog_entry_data, {})
    catalog_entry = CatalogEntry(entry_data, catalog_entry_data, catalog_instance, ["path"])

    # Parse the component entry
    parser = KfpComponentParser.create_instance(platform=RUNTIME_PROCESSOR)
    component = parser.parse(catalog_entry)[0]
    props_as_dict = {param.ref: param for param in component.input_properties}

    # Ensure description with unsafe characters is rendered without error
    palette_json = ComponentCache.to_canvas_palette([component])
    expected_description = 'This component description contains an unescaped " character'
    assert expected_description == palette_json["categories"][0]["node_types"][0]["description"]

    properties_json = ComponentCache.to_canvas_properties(component)

    # Helper method to retrieve the requested parameter value from the dictionary
    def get_parameter_value(param_name):
        property_dict = properties_json["properties"]["component_parameters"]["properties"][param_name]
        return property_dict["oneOf"][0]["properties"]["value"].get("default", "")

    # Helper method to retrieve the requested parameter info from the dictionary
    def get_parameter_format(param_name):
        property_dict = properties_json["properties"]["component_parameters"]["properties"][param_name]
        return property_dict["oneOf"][0]["properties"]["value"]["type"]

    # Helper method to retrieve the requested parameter description from the dictionary
    def get_parameter_description(param_name):
        property_dict = properties_json["properties"]["component_parameters"]["properties"][param_name]
        return property_dict["description"]

    # Helper method to retrieve whether the requested parameter is required
    def get_parameter_required(param_name):
        required_parameters = properties_json["properties"]["component_parameters"]["required"]
        return param_name in required_parameters

    # Ensure system parameters are present
    assert properties_json["properties"]["label"] is not None

    component_source = json.dumps({"catalog_type": catalog_type, "component_ref": catalog_entry.entry_reference})
    assert properties_json["properties"]["component_source"]["default"] == component_source

    assert get_parameter_value("test_string_no_default") == ""
    assert get_parameter_value("test_string_default_value") == "default"
    assert get_parameter_value("test_string_default_empty") == ""
    assert get_parameter_value("test_bool_default") is False
    assert get_parameter_value("test_bool_false") is False
    assert get_parameter_value("test_bool_true") is True
    assert get_parameter_value("test_int_default") == 0
    assert get_parameter_value("test_int_zero") == 0
    assert get_parameter_value("test_int_non_zero") == 1
    assert get_parameter_value("test_float_default") == 0.0
    assert get_parameter_value("test_float_zero") == 0.0
    assert get_parameter_value("test_float_non_zero") == 1.1
    assert get_parameter_value("test_dict_default") == "{}"
    assert get_parameter_value("test_list_default") == "[]"
    assert get_parameter_value("mounted_volumes") == ""

    # Ensure that the 'required' attribute was set correctly. KFP components default to required
    # unless explicitly marked otherwise in component YAML.
    assert get_parameter_required("test_required_property") is True
    assert get_parameter_required("test_optional_property") is False
    assert get_parameter_required("test_required_property_default") is True

    # Ensure that type information is inferred correctly
    assert get_parameter_format("test_unusual_type_dict") == "string"
    assert props_as_dict["test_unusual_type_dict"].json_data_type == "object"

    assert get_parameter_format("test_unusual_type_list") == "string"
    assert props_as_dict["test_unusual_type_list"].json_data_type == "array"

    assert get_parameter_format("test_unusual_type_string") == "string"
    assert props_as_dict["test_unusual_type_string"].json_data_type == "string"

    assert get_parameter_format("test_unusual_type_notgiven") == "string"
    assert props_as_dict["test_unusual_type_notgiven"].json_data_type == "string"

    # Ensure descriptions are rendered properly with type hint in parentheses
    assert (
        get_parameter_description("test_unusual_type_dict") == "The test command description "
        "(type: Dictionary of arrays)"
    )
    assert get_parameter_description("test_unusual_type_list") == "The test command description (type: An array)"
    assert get_parameter_description("test_unusual_type_string") == "The test command description (type: A string)"
    assert get_parameter_description("test_unusual_type_notgiven") == "The test command description (type: String)"


def test_parse_kfp_component_url():
    # Define the appropriate reader for a URL-type component definition
    kfp_supported_file_types = [".yaml"]
    reader = UrlComponentCatalogConnector(kfp_supported_file_types)

    # Read contents of given path
    url = "https://raw.githubusercontent.com/kubeflow/pipelines/sdk-2.8.0/components/contrib/notebooks/Run_notebook_using_papermill/component.yaml"  # noqa: E501
    catalog_entry_data = {"url": url}

    # Construct a catalog instance
    catalog_type = "url-catalog"
    catalog_instance = ComponentCatalogMetadata(
        schema_name=catalog_type, metadata={"categories": ["Test"], "runtime_type": RUNTIME_PROCESSOR.name}
    )

    # Build the catalog entry data structures required for parsing
    entry_data = reader.get_entry_data(catalog_entry_data, {})
    catalog_entry = CatalogEntry(entry_data, catalog_entry_data, catalog_instance, ["url"])

    # Parse the component entry
    parser = KfpComponentParser.create_instance(platform=RUNTIME_PROCESSOR)
    component = parser.parse(catalog_entry)[0]

    properties_json = ComponentCache.to_canvas_properties(component)

    # Helper method to retrieve the requested parameter value from the dictionary
    def get_parameter_value(param_name):
        property_dict = properties_json["properties"]["component_parameters"]["properties"][param_name]
        if "oneOf" not in property_dict:
            # This is an inputpath-only param and doesn't have a default value
            return None
        return property_dict["oneOf"][0]["properties"]["value"].get("default", "")

    # Ensure system parameters are present
    assert properties_json["properties"]["label"] is not None

    component_source = json.dumps({"catalog_type": catalog_type, "component_ref": catalog_entry.entry_reference})
    assert properties_json["properties"]["component_source"]["default"] == component_source

    assert get_parameter_value("notebook") is None  # Default value for type `inputpath`
    assert get_parameter_value("parameters") == "{}"
    assert get_parameter_value("packages_to_install") == "[]"
    assert get_parameter_value("input_data") == ""

    # Ensure 'Notebook' param exists as both an input and an output (same name defined in yaml)
    assert "notebook" in properties_json["properties"]["component_parameters"]["properties"]
    assert "output_notebook" in properties_json["properties"]["component_parameters"]["properties"]


def test_parse_kfp_component_file_no_inputs():
    # Define the appropriate reader for a filesystem-type component definition
    kfp_supported_file_types = [".yaml"]
    reader = FilesystemComponentCatalogConnector(kfp_supported_file_types)

    # Read contents of given path
    path = _get_resource_path("kfp_test_operator_no_inputs.yaml")
    catalog_entry_data = {"path": path}

    # Construct a catalog instance
    catalog_type = "local-file-catalog"
    catalog_instance = ComponentCatalogMetadata(
        schema_name=catalog_type, metadata={"categories": ["Test"], "runtime_type": RUNTIME_PROCESSOR.name}
    )

    # Build the catalog entry data structures required for parsing
    entry_data = reader.get_entry_data(catalog_entry_data, {})
    catalog_entry = CatalogEntry(entry_data, catalog_entry_data, catalog_instance, ["path"])

    # Parse the component entry
    parser = KfpComponentParser.create_instance(platform=RUNTIME_PROCESSOR)
    component = parser.parse(catalog_entry)[0]

    properties_json = ComponentCache.to_canvas_properties(component)

    # Properties JSON should only include the six parameters common to every
    # component ('mounted_volumes', 'kubernetes_pod_annotations', 'kubernetes_pod_labels',
    # 'kubernetes_tolerations', 'kubernetes_shared_mem_size', and 'disable_node_caching),
    # and the output parameter for this component
    num_common_params = 7
    properties_from_json = [
        prop
        for prop in properties_json["properties"]["component_parameters"]["properties"].keys()
        if "header" not in prop
    ]
    assert len(properties_from_json) == num_common_params

    # Ensure system parameters are present
    assert properties_json["properties"]["label"] is not None

    component_source = json.dumps({"catalog_type": catalog_type, "component_ref": catalog_entry.entry_reference})
    assert properties_json["properties"]["component_source"]["default"] == component_source


async def test_parse_components_not_a_file():
    # Define the appropriate reader for a filesystem-type component definition
    kfp_supported_file_types = [".yaml"]
    reader = FilesystemComponentCatalogConnector(kfp_supported_file_types)

    # Get path to an invalid component definition file and read contents
    path = _get_resource_path("kfp_test_operator_not_a_file.yaml")
    entry_data = reader.get_entry_data({"path": path}, {})
    assert entry_data is None


async def test_parse_components_invalid_yaml(caplog):
    # Get resource path and read definition (by-pass catalog reader functionality)
    path = _get_resource_path("kfp_test_invalid_component.yaml")
    with open(path, "r") as f:
        definition = f.read()

    # Manually construct catalog_entry_data object and catalog instance
    catalog_entry_data = {"path": path}
    catalog_type = "local-file-catalog"
    catalog_instance = ComponentCatalogMetadata(
        schema_name=catalog_type, metadata={"categories": ["Test"], "runtime_type": RUNTIME_PROCESSOR.name}
    )

    # Build the catalog entry data structures required for parsing
    entry_data = EntryData(definition=definition)
    catalog_entry = CatalogEntry(entry_data, catalog_entry_data, catalog_instance, ["path"])

    # Parse the component entry
    parser = KfpComponentParser.create_instance(platform=RUNTIME_PROCESSOR)
    component = parser.parse(catalog_entry)

    # Failed YAML schema validation returns None
    assert component is None

    # Assert validation error is captured appropriately in log
    assert "Invalid format of YAML definition for component" in caplog.text
    assert "Failed validating 'type'" in caplog.text
    assert "On instance['inputs'][0]['name']:\n    2" in caplog.text

    caplog.clear()

    # Modify file to get expected error in YAML safe_load
    new_definition = "key with no mapping\n" + definition
    catalog_entry.entry_data.definition = new_definition

    # Re-parse with new definition content
    component = parser.parse(catalog_entry)

    # Failed YAML safe_load returns None
    assert component is None

    # Assert load error is captured appropriately in log
    assert "Could not load YAML definition for component" in caplog.text
    assert "mapping values are not allowed here" in caplog.text


async def test_parse_components_additional_metatypes():
    # Define the appropriate reader for a URL-type component definition
    kfp_supported_file_types = [".yaml"]
    reader = UrlComponentCatalogConnector(kfp_supported_file_types)

    # Read contents of given path
    url = "https://raw.githubusercontent.com/kubeflow/pipelines/sdk-2.8.0/components/contrib/keras/Train_classifier/from_CSV/component.yaml"  # noqa: E501
    catalog_entry_data = {"url": url}

    # Construct a catalog instance
    catalog_type = "url-catalog"
    catalog_instance = ComponentCatalogMetadata(
        schema_name=catalog_type, metadata={"categories": ["Test"], "runtime_type": RUNTIME_PROCESSOR.name}
    )

    # Build the catalog entry data structures required for parsing
    entry_data = reader.get_entry_data(catalog_entry_data, {})
    catalog_entry = CatalogEntry(entry_data, catalog_entry_data, catalog_instance, ["url"])

    # Parse the component entry
    parser = KfpComponentParser()
    component = parser.parse(catalog_entry)[0]
    properties_json = ComponentCache.to_canvas_properties(component)

    # Helper method to retrieve the requested parameter value from the dictionary
    def get_parameter_value(param_name):
        property_dict = properties_json["properties"]["component_parameters"]["properties"][param_name]
        if "oneOf" not in property_dict:
            # This is an inputpath-only param and doesn't have a default value
            return None
        return property_dict["oneOf"][0]["properties"]["value"].get("default", "")

    # Ensure system parameters are present
    assert properties_json["properties"]["label"] is not None

    component_source = json.dumps({"catalog_type": catalog_type, "component_ref": catalog_entry.entry_reference})
    assert properties_json["properties"]["component_source"]["default"] == component_source

    assert get_parameter_value("training_features") is None  # inputPath
    assert get_parameter_value("training_labels") is None  # inputPath
    assert get_parameter_value("network_json") is None  # inputPath
    assert get_parameter_value("loss_name") == "categorical_crossentropy"
    assert get_parameter_value("num_classes") == 0
    assert get_parameter_value("optimizer") == "rmsprop"
    assert get_parameter_value("optimizer_config") == ""
    assert get_parameter_value("learning_rate") == 0.01
    assert get_parameter_value("num_epochs") == 100
    assert get_parameter_value("batch_size") == 32
    assert get_parameter_value("metrics") == "['accuracy']"
    assert get_parameter_value("random_seed") == 0



#
# Copyright 2018-2023 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Optional

# this may raise an ImportError if the python-gitlab package is not installed
from gitlab import Gitlab

# this may raise an ImportError if the python-gitlab package is not installed
from gitlab.exceptions import GitlabError  # noqa H306
from traitlets.config import LoggingConfigurable
from urllib3.util import parse_url


class GitLabClient(LoggingConfigurable):
    def __init__(
        self,
        token: str,
        project: str,
        branch: Optional[str] = None,
        server_url: Optional[str] = "https://gitlab.com",
        **kwargs,
    ):
        """
        Creates a GitLab client for Elyra
        :param token: Personal Access Token for use with GitLab
        :param project: GitLab project to use. Use format [namespace]/[project] e.g. elyra/examples
        :param branch: Project branch to use. If not provided, this will use the default branch configured in the
                       target project
        :param server_url:  GitLab API endpoint to use for the client. This is can be an Enterprise
                          GitLab instance. By default the client will attempt to connect to the main GitLab API at
                          https://www.gitlab.com'
        """

        super().__init__(**kwargs)

        # Remove trailing slash(es) from server URL to prevent failure
        self.server_url = server_url.rstrip("/")
        self.project_name = project
        self.branch = branch

        try:
            self.client = Gitlab(self.server_url, private_token=token)
            self.gitlab_project = self.client.projects.get(self.project_name)
        except GitlabError as gle:
            self.log.error(f"Error accessing project {self.project_name}: {gle}")
            raise RuntimeError(
                f"Error accessing repository {self.project_name}: {gle}. "
                "Please validate your runtime configuration details and retry."
            ) from gle

    def upload_dag(self, pipeline_filepath: str, pipeline_name: str) -> None:
        """
        Push a DAG to a gitlab project
        :param pipeline_filepath: filepath to the location of the DAG in the local filesystem
        :param pipeline_name: the name of the file to be created in the gitlab project
        :return:
        """
        try:
            # Upload DAG to gitlab project
            with open(pipeline_filepath) as input_file:
                content = input_file.read()

                git_file_name = f"{pipeline_name}.py"

                self.gitlab_project.files.create(
                    {
                        "file_path": git_file_name,
                        "branch": self.branch,
                        "content": content,
                        "commit_message": f"Pushed DAG {pipeline_name}",
                    }
                )

            self.log.info(f"DAG file {git_file_name} was successfully uploaded to branch {self.branch}.")

        except FileNotFoundError as fnfe:
            self.log.error(f"Unable to locate local DAG file to upload: {pipeline_filepath}: " + str(fnfe))
            raise RuntimeError(f"Unable to locate local DAG file to upload: {pipeline_filepath}: {str(fnfe)}") from fnfe
        except GitlabError as gle:
            self.log.error(f"Error uploading DAG to branch {self.branch}: {gle}")
            raise RuntimeError(
                f"Error uploading DAG to branch {self.branch}: {gle} "
                "Please validate your runtime configuration details and try again."
            ) from gle

    @staticmethod
    def get_git_url(api_url: str, repository_name: str, repository_branch: str) -> str:
        """
        Generates the URL to the location of the pushed DAG
        :param api_url: git API endpoint URL
        :param project_name: name of the GitLab project in the form [namespace]/[project]
        :param project_branch: name of the project branch
        :return: a URL in string format
        """
        parsed_url = parse_url(api_url)
        scheme = f"{parsed_url.scheme}://"
        host = parsed_url.host
        port = ""

        if parsed_url.host.split(".")[0] == "api":
            host = ".".join(parsed_url.host.split(".")[1:])

        if parsed_url.port:
            port = f":{parsed_url.port}"

        return f"{scheme}{host}{port}/{repository_name}/tree/{repository_branch}"



#
# Copyright 2018-2023 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys

import pytest

from elyra.pipeline.pipeline import GenericOperation
from elyra.pipeline.pipeline import Operation
from elyra.pipeline.pipeline import Pipeline
from elyra.pipeline.properties import ElyraProperty


@pytest.fixture
def good_operation():
    component_parameters = {
        "filename": "elyra/pipeline/tests/resources/archive/test.ipynb",
        "runtime_image": "tensorflow/tensorflow:latest",
    }
    test_operation = GenericOperation(
        id="test-id",
        type="execution-node",
        classifier="execute-notebook-node",
        name="test",
        component_props=component_parameters,
    )
    return test_operation


@pytest.fixture
def good_pipeline():
    test_pipeline = Pipeline(
        id="Random-UUID-123123123123123", name="test-pipeline", runtime="kfp", runtime_config="default_kfp"
    )
    return test_pipeline


def test_create_operation_minimal(good_operation):
    test_operation = good_operation

    assert test_operation.id == "test-id"
    assert test_operation.type == "execution-node"
    assert test_operation.classifier == "execute-notebook-node"
    assert test_operation.name == "test"
    assert test_operation.filename == "elyra/pipeline/tests/resources/archive/test.ipynb"
    assert test_operation.runtime_image == "tensorflow/tensorflow:latest"
    assert test_operation.name == "test"


def test_create_operation_with_dependencies():
    dependencies = ["elyra/pipline/tests/resources", "elyra/pipline/tests/resources/archive"]

    component_parameters = {
        "filename": "elyra/pipeline/tests/resources/archive/test.ipynb",
        "dependencies": dependencies,
        "runtime_image": "tensorflow/tensorflow:latest",
    }
    test_operation = GenericOperation(
        id="test-id",
        type="execution-node",
        classifier="execute-notebook-node",
        name="test",
        component_props=component_parameters,
    )

    assert test_operation.dependencies == dependencies


def test_create_operation_include_subdirectories():
    include_subdirectories = True

    component_parameters = {
        "filename": "elyra/pipeline/tests/resources/archive/test.ipynb",
        "include_subdirectories": include_subdirectories,
        "runtime_image": "tensorflow/tensorflow:latest",
    }
    test_operation = GenericOperation(
        id="test-id",
        type="execution-node",
        classifier="execute-notebook-node",
        name="test",
        component_props=component_parameters,
    )

    assert test_operation.include_subdirectories == include_subdirectories


def test_create_operation_with_environmental_variables():
    env_variables = ['FOO="Bar"', 'BAR="Foo']

    component_parameters = {
        "filename": "elyra/pipeline/tests/resources/archive/test.ipynb",
        "runtime_image": "tensorflow/tensorflow:latest",
    }
    test_operation = GenericOperation(
        id="test-id",
        type="execution-node",
        classifier="execute-notebook-node",
        name="test",
        component_props=component_parameters,
        elyra_props={"env_vars": env_variables},
    )

    assert test_operation.env_vars == env_variables


def test_create_operation_with_inputs():
    inputs = ["input1.txt", "input2.txt"]

    component_parameters = {
        "filename": "elyra/pipeline/tests/resources/archive/test.ipynb",
        "inputs": inputs,
        "runtime_image": "tensorflow/tensorflow:latest",
    }
    test_operation = GenericOperation(
        id="test-id",
        type="execution-node",
        classifier="execute-notebook-node",
        name="test",
        component_props=component_parameters,
    )

    assert test_operation.inputs == inputs


def test_create_operation_with_outputs():
    outputs = ["output1.txt", "output2.txt"]

    component_parameters = {
        "filename": "elyra/pipeline/tests/resources/archive/test.ipynb",
        "outputs": outputs,
        "runtime_image": "tensorflow/tensorflow:latest",
    }
    test_operation = GenericOperation(
        id="test-id",
        type="execution-node",
        classifier="execute-notebook-node",
        name="test",
        component_props=component_parameters,
    )

    assert test_operation.outputs == outputs


def test_create_operation_with_parent_operations():
    parent_operation_ids = ["id-123123-123123-123123", "id-456456-456456-456456"]

    component_parameters = {
        "filename": "elyra/pipeline/tests/resources/archive/test.ipynb",
        "runtime_image": "tensorflow/tensorflow:latest",
    }
    test_operation = GenericOperation(
        id="test-id",
        type="execution-node",
        classifier="execute-notebook-node",
        name="test",
        parent_operation_ids=parent_operation_ids,
        component_props=component_parameters,
    )

    assert test_operation.parent_operation_ids == parent_operation_ids


def test_create_operation_correct_naming():
    label = "test.ipynb"
    filename = "elyra/pipeline/tests/resources/archive/" + label

    component_parameters = {"filename": filename, "runtime_image": "tensorflow/tensorflow:latest"}
    test_operation = GenericOperation(
        id="test-id",
        type="execution-node",
        classifier="execute-notebook-node",
        name=label,
        component_props=component_parameters,
    )

    assert test_operation.name == label.split(".")[0]


def test_fail_create_operation_missing_id():
    component_parameters = {
        "filename": "elyra/pipeline/tests/resources/archive/test.ipynb",
        "runtime_image": "tensorflow/tensorflow:latest",
    }
    with pytest.raises(TypeError):
        Operation(
            type="execution-node",
            classifier="execute-notebook-node",
            name="test",
            component_props=component_parameters,
        )


def test_fail_create_operation_missing_type():
    component_parameters = {
        "filename": "elyra/pipeline/tests/resources/archive/test.ipynb",
        "runtime_image": "tensorflow/tensorflow:latest",
    }
    with pytest.raises(TypeError):
        GenericOperation(
            id="test-id", classifier="execute-notebook-node", name="test", component_props=component_parameters
        )


def test_fail_create_operation_missing_classifier():
    component_parameters = {
        "filename": "elyra/pipeline/tests/resources/archive/test.ipynb",
        "runtime_image": "tensorflow/tensorflow:latest",
    }
    with pytest.raises(TypeError):
        Operation(id="test-id", type="execution-node", name="test", component_props=component_parameters)


def test_fail_create_operation_missing_runtime_image():
    component_parameters = {"filename": "elyra/pipeline/tests/resources/archive/test.ipynb"}
    with pytest.raises(ValueError):
        GenericOperation(
            id="test-id",
            type="execution-node",
            classifier="execute-notebook-node",
            name="test",
            component_props=component_parameters,
        )


def test_fail_create_operation_missing_name():
    component_parameters = {
        "filename": "elyra/pipeline/tests/resources/archive/test.ipynb",
        "runtime_image": "tensorflow/tensorflow:latest",
    }
    with pytest.raises(TypeError):
        GenericOperation(
            id="test-id",
            type="execution-node",
            classifier="execute-notebook-node",
            component_props=component_parameters,
        )


def test_fail_operations_are_equal(good_operation):
    parent_operation_ids = ["id-123123-123123-123123", "id-456456-456456-456456"]
    component_parameters = {
        "filename": "elyra/pipeline/tests/resources/archive/test.ipynb",
        "runtime_image": "tensorflow/tensorflow:latest",
    }
    compare_operation = GenericOperation(
        id="test-id",
        type="execution-node",
        classifier="execute-notebook-node",
        name="test",
        parent_operation_ids=parent_operation_ids,
        component_props=component_parameters,
    )
    with pytest.raises(AssertionError):
        assert compare_operation == good_operation


def test_operations_are_equal(good_operation):
    component_parameters = {
        "filename": "elyra/pipeline/tests/resources/archive/test.ipynb",
        "runtime_image": "tensorflow/tensorflow:latest",
    }
    compare_operation = GenericOperation(
        id="test-id",
        type="execution-node",
        classifier="execute-notebook-node",
        name="test",
        component_props=component_parameters,
    )

    assert compare_operation == good_operation


def test_create_pipeline_minimal(good_pipeline):
    test_pipeline = good_pipeline

    assert test_pipeline.id == "Random-UUID-123123123123123"
    assert test_pipeline.name == "test-pipeline"
    assert test_pipeline.runtime == "kfp"
    assert test_pipeline.runtime_config == "default_kfp"


def test_fail_create_pipeline_missing_id():
    with pytest.raises(TypeError):
        Pipeline(name="test-pipeline", runtime="kfp", runtime_config="default_kfp")


def test_fail_create_pipeline_missing_name():
    with pytest.raises(TypeError):
        Pipeline(id="Random-UUID-123123123123123", runtime="kfp", runtime_config="default_kfp")


def test_fail_create_pipeline_missing_runtime():
    with pytest.raises(TypeError):
        Pipeline(id="Random-UUID-123123123123123", name="test-pipeline", runtime_config="default_kfp")


def test_pipelines_are_equal(good_pipeline):
    compare_pipeline = Pipeline(
        id="Random-UUID-123123123123123", name="test-pipeline", runtime="kfp", runtime_config="default_kfp"
    )

    assert compare_pipeline == good_pipeline


def test_fail_pipelines_are_equal(good_pipeline):
    test_operations_dict = {"123123123": good_operation, "234234234": good_operation}

    compare_pipeline = Pipeline(
        id="Random-UUID-123123123123123", name="test-pipeline", runtime="kfp", runtime_config="default_kfp"
    )

    for key, operation in test_operations_dict.items():
        compare_pipeline.operations[key] = operation

    with pytest.raises(AssertionError):
        assert compare_pipeline == good_pipeline


def test_env_list_to_dict_function():
    env_variables_dict = {"KEY": "val", "KEY2": "value2", "TWO_EQUALS": "KEY=value", "": "no_key"}
    env_variables = [
        {"env_var": "KEY", "value": "val"},  # valid
        {"env_var": "", "value": ""},  # empty key and value
        {"env_var": "  ", "value": "empty_key"},  # empty key
        {"env_var": "", "value": "no_key"},  # empty key with value
        {"env_var": "EMPTY_VALUE", "value": "  "},  # empty value
        {"env_var": "NO_VALUE"},  # no value
        {"env_var": "KEY2", "value": "value2"},  # valid
        {"env_var": "TWO_EQUALS", "value": "KEY=value"},  # valid
        {},  # no values
        None,  # None value
    ]

    converted_list = ElyraProperty.create_instance("env_vars", env_variables)
    assert converted_list.to_dict() == env_variables_dict


def test_validate_resource_values():
    component_parameters = {
        "filename": "elyra/pipeline/tests/resources/archive/test.ipynb",
        "cpu": "4",
        "memory": "10",
        "gpu": "6",
        "gpu_vendor": "example.com/gpu",
        "runtime_image": "tensorflow/tensorflow:latest",
    }
    test_operation = GenericOperation(
        id="test-id",
        type="execution-node",
        classifier="execute-notebook-node",
        name="test",
        component_props=component_parameters,
    )

    assert test_operation.cpu == "4"
    assert test_operation.gpu == "6"
    assert test_operation.gpu_vendor == "example.com/gpu"
    assert test_operation.memory == "10"


def test_validate_resource_values_as_none():
    component_parameters = {
        "filename": "elyra/pipeline/tests/resources/archive/test.ipynb",
        "runtime_image": "tensorflow/tensorflow:latest",
    }
    test_operation = GenericOperation(
        id="test-id",
        type="execution-node",
        classifier="execute-notebook-node",
        name="test",
        component_props=component_parameters,
    )

    assert test_operation.cpu is None
    assert test_operation.memory is None
    assert test_operation.gpu is None
    assert test_operation.gpu_vendor is None


def test_validate_gpu_accepts_zero_as_value():
    component_parameters = {
        "filename": "elyra/pipeline/tests/resources/archive/test.ipynb",
        "cpu": "4",
        "gpu": "0",
        "memory": "10",
        "runtime_image": "tensorflow/tensorflow:latest",
    }
    test_operation = GenericOperation(
        id="test-id",
        type="execution-node",
        classifier="execute-notebook-node",
        name="test",
        component_props=component_parameters,
    )

    assert test_operation.gpu == "0"


def test_validate_max_resource_value():
    system_max_size = str(sys.maxsize - 1)

    component_parameters = {
        "filename": "elyra/pipeline/tests/resources/archive/test.ipynb",
        "memory": system_max_size,
        "runtime_image": "tensorflow/tensorflow:latest",
    }
    test_operation = GenericOperation(
        id="test-id",
        type="execution-node",
        classifier="execute-notebook-node",
        name="test",
        component_props=component_parameters,
    )

    assert test_operation.memory == system_max_size


def test_fail_validate_max_resource_value_exceeded():
    system_max_size = str(sys.maxsize)

    component_parameters = {
        "filename": "elyra/pipeline/tests/resources/archive/test.ipynb",
        "memory": system_max_size,
        "runtime_image": "tensorflow/tensorflow:latest",
    }
    with pytest.raises(ValueError):
        GenericOperation(
            id="test-id",
            type="execution-node",
            classifier="execute-notebook-node",
            name="test",
            component_props=component_parameters,
        )


def test_fail_creating_operation_with_negative_gpu_resources():
    component_parameters = {
        "filename": "elyra/pipeline/tests/resources/archive/test.ipynb",
        "gpu": "-1",
        "runtime_image": "tensorflow/tensorflow:latest",
    }
    with pytest.raises(ValueError):
        GenericOperation(
            id="test-id",
            type="execution-node",
            classifier="execute-notebook-node",
            name="test",
            component_props=component_parameters,
        )


def test_fail_creating_operation_with_0_cpu_resources():
    component_parameters = {
        "filename": "elyra/pipeline/tests/resources/archive/test.ipynb",
        "cpu": "0",
        "runtime_image": "tensorflow/tensorflow:latest",
    }
    with pytest.raises(ValueError):
        GenericOperation(
            id="test-id",
            type="execution-node",
            classifier="execute-notebook-node",
            name="test",
            component_props=component_parameters,
        )


def test_fail_creating_operation_with_negative_cpu_resources():
    component_parameters = {
        "filename": "elyra/pipeline/tests/resources/archive/test.ipynb",
        "cpu": "-1",
        "runtime_image": "tensorflow/tensorflow:latest",
    }
    with pytest.raises(ValueError):
        GenericOperation(
            id="test-id",
            type="execution-node",
            classifier="execute-notebook-node",
            name="test",
            component_props=component_parameters,
        )


def test_fail_creating_operation_with_more_requests_than_limit_cpu():
    component_parameters = {
        "filename": "elyra/pipeline/tests/resources/archive/test.ipynb",
        "cpu": "4",
        "cpu_limit": "2",
        "runtime_image": "tensorflow/tensorflow:latest",
    }
    with pytest.raises(ValueError):
        GenericOperation(
            id="test-id",
            type="execution-node",
            classifier="execute-notebook-node",
            name="test",
            component_props=component_parameters,
        )


def test_fail_creating_operation_with_more_requests_than_limit_memory():
    component_parameters = {
        "filename": "elyra/pipeline/tests/resources/archive/test.ipynb",
        "memory": "4",
        "memory_limit": "2",
        "runtime_image": "tensorflow/tensorflow:latest",
    }
    with pytest.raises(ValueError):
        GenericOperation(
            id="test-id",
            type="execution-node",
            classifier="execute-notebook-node",
            name="test",
            component_props=component_parameters,
        )


def test_fail_creating_operation_with_0_memory_resources():
    component_parameters = {
        "filename": "elyra/pipeline/tests/resources/archive/test.ipynb",
        "memory": "0",
        "runtime_image": "tensorflow/tensorflow:latest",
    }
    with pytest.raises(ValueError):
        GenericOperation(
            id="test-id",
            type="execution-node",
            classifier="execute-notebook-node",
            name="test",
            component_props=component_parameters,
        )


def test_fail_creating_operation_with_negative_memory_resources():
    component_parameters = {
        "filename": "elyra/pipeline/tests/resources/archive/test.ipynb",
        "memory": "-1",
        "runtime_image": "tensorflow/tensorflow:latest",
    }
    with pytest.raises(ValueError):
        GenericOperation(
            id="test-id",
            type="execution-node",
            classifier="execute-notebook-node",
            name="test",
            component_props=component_parameters,
        )


def test_scrub_list_function():
    env_variables_input = ["FOO=Bar", "BAR=Foo", None, ""]
    env_variables_output = ["FOO=Bar", "BAR=Foo"]

    assert Operation._scrub_list(env_variables_input) == env_variables_output



#
# Copyright 2018-2023 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import pytest

from elyra.metadata.error import MetadataNotFoundError
from elyra.metadata.manager import MetadataManager
from elyra.metadata.metadata import Metadata
from elyra.metadata.schemaspaces import RuntimeImages
from elyra.metadata.schemaspaces import Runtimes
from elyra.metadata.storage import FileMetadataStore
from elyra.pipeline.kfp.processor_kfp import WorkflowEngineType
from elyra.pipeline.parser import PipelineParser
from elyra.pipeline.pipeline import Pipeline
from elyra.pipeline.pipeline_constants import COS_OBJECT_PREFIX


@pytest.fixture()
def metadata_managers(jp_environ, jp_data_dir):
    """
    This fixture creates an elyra.metadata.manager.MetadataManager instance for the
    Runtimes.RUNTIMES_IMAGES_SCHEMASPACE_NAME and an elyra.metadata.manager.MetadataManager instance for the
    Runtimes.RUNTIMES_SCHEMASPACE_NAME.
    The fixture yields a dictionary:
        {
            RuntimeImages.RUNTIMES_IMAGES_SCHEMASPACE_NAME: MetadataManager,
            Runtimes.RUNTIMES_SCHEMASPACE_NAME: MetadataManager
        }
    """
    # Set up the required directory structure
    metadata_dir = Path(jp_data_dir) / "metadata"
    metadata_dir.mkdir(parents=False)
    (metadata_dir / Runtimes.RUNTIMES_SCHEMASPACE_NAME).mkdir()
    (metadata_dir / RuntimeImages.RUNTIMES_IMAGES_SCHEMASPACE_NAME).mkdir()

    # Create MM for Runtimes.RUNTIMES_SCHEMASPACE_NAME
    r_metadata_mgr = MetadataManager(
        schemaspace=Runtimes.RUNTIMES_SCHEMASPACE_NAME, metadata_store_class=FileMetadataStore
    )
    # Create MM for RuntimeImages.RUNTIMES_IMAGES_SCHEMASPACE_NAME
    rti_metadata_mgr = MetadataManager(
        schemaspace=RuntimeImages.RUNTIMES_IMAGES_SCHEMASPACE_NAME, metadata_store_class=FileMetadataStore
    )

    yield {
        RuntimeImages.RUNTIMES_IMAGES_SCHEMASPACE_NAME: rti_metadata_mgr,
        Runtimes.RUNTIMES_SCHEMASPACE_NAME: r_metadata_mgr,
    }


@pytest.fixture()
def rti_mmanager(metadata_managers):
    """
    This fixture creates an elyra.metadata.manager.MetadataManager instance for the
    Runtimes.RUNTIMES_IMAGES_SCHEMASPACE_NAME.
    """
    yield metadata_managers.get(RuntimeImages.RUNTIMES_IMAGES_SCHEMASPACE_NAME)


@pytest.fixture()
def rt_mmanager(metadata_managers):
    """
    This fixture creates an elyra.metadata.manager.MetadataManager instance for the
    Runtimes.RUNTIMES_SCHEMASPACE_NAME.
    """
    yield metadata_managers.get(Runtimes.RUNTIMES_SCHEMASPACE_NAME)


@pytest.fixture()
def runtime_config(rt_mmanager, request):
    """
    Create a persisted Kubeflow Pipelines runtime configuration.
    Optional inputs:
        - "config_name": str (defaults to "test-config")
        - "workflow_engine": WorkflowEngineType.XXX (defaults to WorkflowEngineType.ARGO)
        - "use_cos_credentials_secret": bool,  # (defaults to False)
    The fixture yields the following:
        - elyra.metadata.metadata.Metadata instance
    }
    """
    # Invoke helper method to create a persisted runtime configuration
    yield create_runtime_config(metadata_managers[Runtimes.RUNTIMES_SCHEMASPACE_NAME], request.param)


@pytest.fixture()
def runtime_image_config(rti_mmanager, request):
    """
    Create a persisted runtime image configuration.
    Optional inputs:
        - "image_name": str (defaults to "mocked-container-image")
        - "require_pull_secret": bool (default: False)
    The fixture yields the following:
        - elyra.metadata.metadata.Metadata instance
    """
    image_name = request.param.pop("image_name", "mocked-container-image")
    customization_options = request.param
    # Invoke helper method to create a persisted runtime image configuration
    yield create_runtime_image_config(
        rti_mmanager,
        image_name,
        customization_options,
    )


@pytest.fixture()
def runtime_image_configs(rti_mmanager, request) -> List[Metadata]:
    """
    Create persisted runtime image configurations for the specified pipeline file.
    Required inputs:
     - "pipeline_file": existing pipeline filename
    Optional inputs:
        - "require_pull_secret": bool (default: False)
    The fixture yields the following:
        - "runtime_image_configs": array of elyra.metadata.metadata.Metadata instances
    """
    # check required parameters
    assert request.param.get("pipeline_file") is not None, "A pipeline filename is required."

    # Load pipeline
    # Create runtime image configurations
    customization_options = {
        "require_pull_secret": request.param.get("require_pull_secret"),
    }
    # Invoke helper method to create persisted runtime image configurations
    yield create_runtime_image_configs(
        rti_metadata_manager=metadata_managers[RuntimeImages.RUNTIMES_IMAGES_SCHEMASPACE_NAME],
        pipeline_file=request.param.get("pipeline_file"),
        customization_options=customization_options,
    )


@pytest.fixture()
def metadata_dependencies(metadata_managers, request):
    """
    Create Elyra metadata repository artifacts that are required by the specified
    pipeline: a runtime configuration and zero or more runtime image configurations.
    Required inputs:
     - "pipeline_file": existing pipeline filename
    Optional inputs:
        - "with_cos_object_prefix": bool (default: False)
        - "workflow_engine": WorkflowEngineType.ARGO or WorkflowEngineType.TEKTON
        - "use_cos_credentials_secret": bool (default: False)
        - "require_pull_secret": bool (default: False)
    The fixture yields a dictionary with the following keys:
     - "pipeline_file": the pipeline file name, as specified as input
     - "pipeline_object": elyra.pipeline.pipeline.Pipeline instance
     - "runtime_config": elyra.metadata.metadata.Metadata instance
     - "runtime_image_configs": array of elyra.metadata.metadata.Metadata instances
     - "metadata_managers": dictionary [str, elyra.metadata.manager.MetadataManager]
       (keys are RuntimeImages.RUNTIMES_IMAGES_SCHEMASPACE_NAME and
       Runtimes.RUNTIMES_SCHEMASPACE_NAME)
     - "fixture_parameters": dictionary containing the fixture's inputs
    """
    # check required parameters
    assert request.param.get("pipeline_file") is not None, "A pipeline filename is required."

    # Create a Pipeline object from the pipeline file, applying the customization
    # options
    customization_options = {}
    for supported_option in [
        "with_cos_object_prefix",
        "resources_cpu",
        "resources_cpu_limit",
        "resources_gpu",
        "resources_memory",
        "resources_memory_limit",
    ]:
        customization_options[supported_option] = request.param.get(supported_option)
    pipeline_object = get_pipeline_object(
        pipeline_filename=request.param["pipeline_file"],
        customization_options=customization_options,
    )

    # Create runtime configuration
    runtime_config_options = {
        "config_name": pipeline_object.runtime_config,
        "workflow_engine": request.param.get("workflow_engine"),
        "use_cos_credentials_secret": request.param.get("use_cos_credentials_secret"),
    }
    runtime_config = create_runtime_config(
        metadata_managers[Runtimes.RUNTIMES_SCHEMASPACE_NAME], runtime_config_options
    )

    # Create runtime image configurations
    runtime_image_options = {"require_pull_secret": request.param.get("require_pull_secret")}
    runtime_image_configs = create_runtime_image_configs(
        metadata_managers[RuntimeImages.RUNTIMES_IMAGES_SCHEMASPACE_NAME],
        pipeline_file=request.param["pipeline_file"],
        pipeline_object=pipeline_object,
        customization_options=runtime_image_options,
    )

    yield {
        "pipeline_file": request.param["pipeline_file"],
        "pipeline_object": pipeline_object,
        "runtime_config": runtime_config,
        "runtime_image_configs": runtime_image_configs,
        "metadata_managers": metadata_managers,
        "fixture_parameters": request.param,
    }


# ------------------------------
# Supporting functions
# ------------------------------


def create_runtime_config(rt_metadata_manager: MetadataManager, customization_options: Dict[str, Any] = {}) -> Metadata:
    """
    Create a persisted Kubeflow Pipelines runtime configuration. The generated
    runtime configuration is customizable using these options properties:
        - "config_name": "custom-config-name",  # default is "test-config"
        - "workflow_engine": WorkflowEngineType.XXX,  # Enum value, ARGO is default
        - "use_cos_credentials_secret": True,  # or False (default)
    """
    kfp_runtime_config = {
        "display_name": "Test runtime configuration",
        "schema_name": "kfp",
        "metadata": {
            "display_name": "Test runtime configuration",
            "runtime_type": "KUBEFLOW_PIPELINES",
            "api_endpoint": "http://examples.com:31737",
            "user_namespace": "kubeflow-user-example-com",
            "auth_type": "DEX_STATIC_PASSWORDS",
            "api_username": "user@example.com",
            "api_password": "12341234",
            "cos_endpoint": "http://examples.com:31671",
            "cos_bucket": "test",
            "tags": ["test"],
        },
    }

    if customization_options.get("workflow_engine") == WorkflowEngineType.TEKTON:
        kfp_runtime_config["metadata"]["engine"] = "Tekton"
    else:
        kfp_runtime_config["metadata"]["engine"] = "Argo"

    if customization_options.get("use_cos_credentials_secret"):
        kfp_runtime_config["metadata"]["cos_auth_type"] = "KUBERNETES_SECRET"
        kfp_runtime_config["metadata"]["cos_username"] = "my_name"
        kfp_runtime_config["metadata"]["cos_password"] = "my_password"
        kfp_runtime_config["metadata"]["cos_secret"] = "secret-name"
    else:
        kfp_runtime_config["metadata"]["cos_auth_type"] = "USER_CREDENTIALS"
        kfp_runtime_config["metadata"]["cos_username"] = "my_name"
        kfp_runtime_config["metadata"]["cos_password"] = "my_password"

    metadata = Metadata(
        name=customization_options.get("config_name", "test-config"),
        display_name=kfp_runtime_config["display_name"],
        schema_name=kfp_runtime_config["schema_name"],
        metadata=kfp_runtime_config["metadata"],
    )
    # persist and return runtime configuration
    return rt_metadata_manager.create(metadata.name, metadata)


def create_runtime_image_config(
    rti_metadata_manager: MetadataManager,
    image_name: str = "mocked-container-image",
    customization_options: Dict[str, Any] = {},
) -> Metadata:
    """
    Create a persisted runtime image configuration for the specified container
    image. The generated runtime image is customizable using these options:
        - "require_pull_secret": True/False,  # default is False
    """
    if not image_name:
        image_name = "mocked-container-image"

    # Generate a metadata filename using the provided container image name as input.
    config_name = sanitize_container_image_name(image_name)

    update_required = True
    try:
        rti_metadata_manager.get(config_name)
    except MetadataNotFoundError:
        update_required = False

    m = {
        "image_name": image_name,
        "pull_policy": "IfNotPresent",
    }

    if customization_options.get("require_pull_secret"):
        m["pull_secret"] = f"{sanitize_container_image_name(image_name)}-secret"

    rti_metadata = Metadata(
        name=config_name,
        display_name=f"Container image {image_name}",
        schema_name="runtime-image",
        metadata=m,
    )
    # persist runtime image configuration
    if update_required:
        persisted_metadata = rti_metadata_manager.update(rti_metadata.name, rti_metadata)
    else:
        persisted_metadata = rti_metadata_manager.create(rti_metadata.name, rti_metadata)

    # return persisted runtime image configuration
    return persisted_metadata


def create_runtime_image_configs(
    rti_metadata_manager: MetadataManager,
    pipeline_file: Union[Path, str],
    pipeline_object: Pipeline = None,
    customization_options: Dict[str, Any] = {},
) -> List[Metadata]:
    """
    Create persisted runtime image configurations for the specified pipeline_file
    or pipeline_object, applying the specified customization options:
      - "require_pull_secret" (True/False [default])
    """
    assert rti_metadata_manager is not None

    # Create pipeline object, if one wasn't provided
    if pipeline_object is None:
        pipeline_object = get_pipeline_object(
            pipeline_filename=pipeline_file, customization_options=customization_options
        )
    # For each generic node identify its runtime image
    images = {}
    for op in pipeline_object.operations.values():
        if op.is_generic:
            images[op.runtime_image] = {}
            if customization_options.get("require_pull_secret"):
                images[op.runtime_image]["require_pull_secret"] = True
    # Generate runtime image configurations for each unique image,
    # applying the provided customization_options
    runtime_image_configs = []
    for image_name, config_options in images.items():
        runtime_image_configs.append(
            create_runtime_image_config(
                rti_metadata_manager=rti_metadata_manager, image_name=image_name, customization_options=config_options
            )
        )
    return runtime_image_configs


def get_pipeline_object(
    pipeline_filename: Union[Path, str], customization_options: Optional[Dict[str, Any]]
) -> Pipeline:
    """
    Creates a KFP Pipeline instance from pipeline_filename, taking into account the
    following optional customization options:
      - "with_cos_object_prefix" (True/False)
      - "resources_cpu" (number, applied to all generic nodes)
      - "resources_gpu" (number, applied to all generic nodes)
      - "resources_memory" (number, applied to all generic nodes)
      - "resources_cpu_limit" (number, applied to all generic nodes)
      - "resources_memory_limit" (number, applied to all generic nodes)
    """
    assert pipeline_filename is not None, "A pipeline filename is required."

    if not isinstance(pipeline_filename, Path):
        pipeline_filename = Path(pipeline_filename)

    assert pipeline_filename.is_file(), f"Pipeline '{pipeline_filename}' does not exist."

    # Load pipeline file content
    with open(pipeline_filename, "r") as fh:
        pipeline_json = json.loads(fh.read())

    # This rudimentary implementation assumes that the provided file is a valid
    # pipeline file, which contains a primary pipeline.
    if len(pipeline_json["pipelines"]) > 0:
        primary_pipeline = pipeline_json["pipelines"][0]
        app_data = primary_pipeline["app_data"]
        # Add runtime information
        if app_data.get("runtime", None) is None:
            app_data["runtime"] = "Kubeflow Pipelines"
        if app_data.get("runtime_type", None) is None:
            app_data["runtime_type"] = "KUBEFLOW_PIPELINES"
        # Add the filename as pipeline source information
        if app_data.get("source", None) is None:
            app_data["source"] = pipeline_filename.name
        # Add runtime configuration name
        if app_data.get("runtime_config", None) is None:
            app_data["runtime_config"] = "test-config"

        if customization_options.get("with_cos_object_prefix"):
            # Define a dummy COS prefix, if none is defined
            if app_data["properties"].get("pipeline_defaults") is None:
                app_data["properties"]["pipeline_defaults"] = {}
            if app_data["properties"]["pipeline_defaults"].get(COS_OBJECT_PREFIX) is None:
                app_data["properties"]["pipeline_defaults"][COS_OBJECT_PREFIX] = "test/project"
        else:
            # Remove the prefix, if one is already defined
            if app_data["properties"].get("pipeline_defaults") is not None:
                app_data["properties"]["pipeline_defaults"].pop(COS_OBJECT_PREFIX, None)

        #
        # Add resource customizations to every generic node
        for node in primary_pipeline["nodes"]:
            if node["op"] == "execute-notebook-node":
                if customization_options.get("resources_cpu") is not None:
                    node["app_data"]["component_parameters"]["cpu"] = customization_options["resources_cpu"]
                else:
                    node["app_data"]["component_parameters"].pop("cpu", None)
                if customization_options.get("resources_cpu_limit") is not None:
                    node["app_data"]["component_parameters"]["cpu_limit"] = customization_options["resources_cpu_limit"]
                else:
                    node["app_data"]["component_parameters"].pop("cpu_limit", None)
                if customization_options.get("resources_gpu") is not None:
                    node["app_data"]["component_parameters"]["gpu"] = customization_options["resources_gpu"]
                else:
                    node["app_data"]["component_parameters"].pop("gpu", None)
                if customization_options.get("resources_memory") is not None:
                    node["app_data"]["component_parameters"]["memory"] = customization_options["resources_memory"]
                else:
                    node["app_data"]["component_parameters"].pop("memory", None)
                if customization_options.get("resources_memory_limit") is not None:
                    node["app_data"]["component_parameters"]["memory_limit"] = customization_options[
                        "resources_memory_limit"
                    ]
                else:
                    node["app_data"]["component_parameters"].pop("memory_limit", None)

    # Parse JSON and return Pipeline instance
    return PipelineParser().parse(pipeline_json=pipeline_json)


def sanitize_container_image_name(name: str) -> str:
    """
    This helper function sanitizes the provided container image name (e.g.
    'quay.io/myorg/my-image:my-tag') by replacing the following characters
    to a dash ('-'): space, colon, forward slash, and dot.
    """
    safe_char = "-"
    return name.lower().replace(" ", safe_char).replace(":", safe_char).replace("/", safe_char).replace(".", safe_char)



#
# Copyright 2018-2023 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from abc import abstractmethod
from dataclasses import dataclass
from dataclasses import field
from importlib import import_module
import json
from logging import Logger
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from traitlets.config import LoggingConfigurable

from elyra.pipeline.properties import ComponentProperty
from elyra.pipeline.properties import ElyraProperty
from elyra.pipeline.runtime_type import RuntimeProcessorType

# Rather than importing only the CatalogEntry class needed in the Component parse
# type hint below, the catalog_connector module must be imported in its
# entirety in order to avoid a circular reference issue
try:
    from elyra.pipeline import catalog_connector
except ImportError:
    import sys

    catalog_connector = sys.modules[f"{__package__}.catalog_connector"]


class Component(object):
    """
    Represents a generic or runtime-specific component
    """

    def __init__(
        self,
        id: str,
        name: str,
        description: Optional[str],
        catalog_type: str,
        component_reference: Any,
        definition: Optional[str] = None,
        runtime_type: Optional[RuntimeProcessorType] = None,
        op: Optional[str] = None,
        categories: Optional[List[str]] = None,
        properties: Optional[List[ComponentProperty]] = None,
        extensions: Optional[List[str]] = None,
        parameter_refs: Optional[dict] = None,
        package_name: Optional[str] = None,
    ):
        """
        :param id: Unique identifier for a component
        :param name: The name of the component for display
        :param description: The description of the component
        :param catalog_type: Indicates the type of component definition resource
                              location; one of ['url', filename', 'directory]
        :param component_reference: Source information to help locate the component definition
        :param definition: The content of the specification file for this component
        :param runtime_type: The runtime type of the component (e.g. KUBEFLOW_PIPELINES, APACHE_AIRFLOW, etc.)
        :param op: The operation name of the component; used by generic components in rendering the palette
        :param categories: A list of categories that this component belongs to; used to organize component
                           in the palette
        :param properties: The set of properties for the component
        :param extensions: The file extension used by the component
        :param package_name: The fully qualified package name (excluding class name) of the file associated
            with this component
        """

        if not id:
            raise ValueError("Invalid component: Missing field 'id'.")
        if not name:
            raise ValueError("Invalid component: Missing field 'name'.")

        self._id = id
        self._name = name
        self._description = description
        self._catalog_type = catalog_type
        self._component_reference = component_reference

        self._definition = definition
        self._runtime_type = runtime_type
        self._op = op
        self._categories = categories or []
        self._properties = properties

        if not parameter_refs:
            if self._catalog_type == "elyra":
                parameter_refs = {"filehandler": "filename"}
            else:
                parameter_refs = {}

        if self._catalog_type == "elyra" and extensions and not parameter_refs.get("filehandler"):
            Component._log_warning(
                f"Component '{self._id}' specifies extensions '{extensions}' but \
                                   no entry in the 'parameter_ref' dictionary for 'filehandler' and \
                                   cannot participate in drag and drop functionality as a result."
            )

        self._extensions = extensions
        self._parameter_refs = parameter_refs
        self._package_name = package_name

    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def description(self) -> Optional[str]:
        return self._description

    @property
    def catalog_type(self) -> str:
        return self._catalog_type

    @property
    def component_reference(self) -> Any:
        return self._component_reference

    @property
    def component_source(self) -> str:
        """
        Informational property consisting of the catalog type from which
        this component originates and the reference information used to
        locate it within that catalog.
        """
        return json.dumps({"catalog_type": self.catalog_type, "component_ref": self.component_reference})

    @property
    def definition(self) -> str:
        return self._definition

    @property
    def runtime_type(self) -> Optional[RuntimeProcessorType]:
        return self._runtime_type

    @property
    def runtime_type_name(self) -> Optional[str]:
        return self._runtime_type.name if isinstance(self._runtime_type, RuntimeProcessorType) else None

    @property
    def op(self) -> Optional[str]:
        return self._op or self._id

    @property
    def categories(self) -> List[str]:
        return self._categories

    @property
    def properties(self) -> Optional[List[ComponentProperty]]:
        return self._properties

    @property
    def extensions(self) -> Optional[List[str]]:
        return self._extensions

    @property
    def parameter_refs(self) -> dict:
        return self._parameter_refs

    @property
    def import_statement(self) -> Optional[str]:
        return f"from {self._package_name} import {self._name}" if self._package_name else None

    @property
    def input_properties(self) -> List[ComponentProperty]:
        return [prop for prop in self._properties if None not in prop.allowed_input_types]

    @property
    def output_properties(self) -> List[ComponentProperty]:
        return [prop for prop in self._properties if None in prop.allowed_input_types]

    @property
    def required_properties(self) -> List[ComponentProperty]:
        return [prop for prop in self.input_properties if prop.required]

    @property
    def file_extension(self) -> Optional[str]:
        """
        The file extension of the definition file representing this
        Component.
        """
        return self.extensions[0] if self.extensions else None

    @staticmethod
    def _log_warning(msg: str, logger: Optional[Logger] = None):
        if logger:
            logger.warning(msg)
        else:
            print(f"WARNING: {msg}")

    def get_elyra_properties(self) -> List[ComponentProperty]:
        """
        Retrieve the list of Elyra-owned ComponentProperty objects that apply to
        this component, removing any whose id collides with a property parsed from
        the component definition.
        """
        op_type = "generic" if self.component_reference == "elyra" else "custom"
        elyra_props = ElyraProperty.get_classes_for_component_type(op_type, self.runtime_type)
        if self.properties:
            # Remove certain Elyra-owned properties if a component-defined property of the same id is already present
            parsed_property_ids = [prop.ref for prop in self.properties]
            elyra_props = [prop for prop in elyra_props if prop.property_id not in parsed_property_ids]
        return elyra_props


class ComponentParser(LoggingConfigurable):  # ABC
    component_platform: RuntimeProcessorType = None
    _file_types: List[str] = None
    _parser_class_map: Dict[str, str] = {
        "APACHE_AIRFLOW": "elyra.pipeline.airflow.component_parser_airflow:AirflowComponentParser",
        "KUBEFLOW_PIPELINES": "elyra.pipeline.kfp.component_parser_kfp:KfpComponentParser",
    }

    @classmethod
    def create_instance(cls, platform: RuntimeProcessorType) -> ComponentParser:
        """
        Class method that creates the appropriate instance of ComponentParser based on platform type name.
        """
        try:
            module_name, class_name = cls._parser_class_map[platform.name].split(":")
            module = import_module(module_name)
            return getattr(module, class_name)()
        except Exception as e:
            raise RuntimeError(f"Could not get appropriate ComponentParser class: {e}")

    @property
    def file_types(self) -> List[str]:
        return self._file_types

    @abstractmethod
    def parse(self, catalog_entry: catalog_connector.CatalogEntry) -> Optional[List[Component]]:
        """
        Parse a component definition given in the catalog entry and return
        a list of fully-qualified Component objects
        """
        raise NotImplementedError()

    def _format_description(self, description: str, data_type: str) -> str:
        """
        Adds property type information parsed from component specification
        to property description.
        """
        if description:
            return f"{description} (type: {data_type})"
        return f"(type: {data_type})"

    def determine_type_information(self, parsed_type: str) -> "PropertyTypeInfo":
        """
        Takes the type information of a component property as parsed from the component
        specification and returns a new type that is one of several standard options.
        """
        parsed_type_lowered = parsed_type.lower()

        # Determine if this is a "container type"
        # Prefer types that occur in a clause of the form "[type] of ..." (i.e., "container" types)
        # E.g. "a dictionary of key/value pairs" will produce the type "dictionary"
        container_types = ["dictionary", "dict", "set", "list", "array", "arr"]
        for option in container_types:
            if option in parsed_type_lowered:
                data_type = option
                if data_type in ["dict", "dictionary"]:
                    data_type = "object"
                    default_value = {}
                else:  # data_type is one of ['list', 'set', 'array', 'arr']
                    data_type = "array"
                    default_value = []

                # Since we know the type, create our return value and bail
                data_type_info = PropertyTypeInfo(
                    parsed_data=parsed_type_lowered, json_data_type=data_type, default_value=default_value
                )
                break
        else:  # None of the container types were found...
            # Standardize type names
            if any(word in parsed_type_lowered for word in ["str", "string"]):
                data_type_info = PropertyTypeInfo(
                    parsed_data=parsed_type_lowered,
                    json_data_type="string",
                )
            elif any(word in parsed_type_lowered for word in ["int", "integer", "number"]):
                data_type_info = PropertyTypeInfo(
                    parsed_data=parsed_type_lowered, json_data_type="number", default_value=0
                )
            elif any(word in parsed_type_lowered for word in ["float"]):
                data_type_info = PropertyTypeInfo(
                    parsed_data=parsed_type_lowered, json_data_type="number", default_value=0.0
                )
            elif any(word in parsed_type_lowered for word in ["bool", "boolean"]):
                data_type_info = PropertyTypeInfo(
                    parsed_data=parsed_type_lowered, json_data_type="boolean", default_value=False
                )
            else:  # Let this be undetermined. Callers should check for this status and adjust
                data_type_info = PropertyTypeInfo(parsed_data=parsed_type_lowered, undetermined=True)

        from elyra.pipeline.processor import PipelineProcessorManager  # placed here to avoid circular reference

        if PipelineProcessorManager.instance().supports_pipeline_params(runtime_type=self.component_platform):
            data_type_info.allowed_input_types.append("parameter")

        return data_type_info


@dataclass
class PropertyTypeInfo:
    """
    This class is initialized by ComponentParser.determine_type_information() and used by subclass
    implementations to determine the current state of parsing a data-type.

    The instance will indicate whether the base ComponentParser could determine the actual data-type
    via a `True` value in its `undetermined` attribute, in which case subclass implementations
    are advised to attempt further parsing. In such cases, the rest of the attributes of the instance
    will reflect a 'string' data type as that is the most flexible data_type and, hence, the default.

    Allowed input types for the given property defaults to the set of all available input types,
    unless the child method is able to determine that the allowed types must be adjusted,
    e.g. kfp path-based types.
    """

    parsed_data: str
    json_data_type: Optional[str] = "string"
    allowed_input_types: Optional[List[str]] = field(default_factory=lambda: ["inputvalue", "inputpath", "file"])
    default_value: Optional[Any] = ""
    required: Optional[bool] = True
    undetermined: Optional[bool] = False



#
# Copyright 2018-2023 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import copy
import errno
import io
import json
import os
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from traitlets.config import LoggingConfigurable

from elyra.metadata.error import MetadataExistsError
from elyra.metadata.error import MetadataNotFoundError
from elyra.metadata.metadata import Metadata
from elyra.metadata.schema import METADATA_TEST_SCHEMASPACE
from elyra.metadata.schema import METADATA_TEST_SCHEMASPACE_ID
from elyra.metadata.schema import Schemaspace
from elyra.metadata.schema import SchemasProvider
from elyra.metadata.storage import FileMetadataStore
from elyra.metadata.storage import MetadataStore


NON_EXISTENT_SCHEMASPACE_ID = "9ab68f6f-000c-470e-814d-2af59ea0956e"

valid_metadata_json = {
    "schema_name": "metadata-test",
    "display_name": "valid metadata instance",
    "metadata": {
        "uri_test": "http://localhost:31823/v1/models?version=2017-02-13",
        "number_range_test": 8,
        "required_test": "required_value",
    },
}

valid_metadata2_json = {
    "schema_name": "metadata-test2",
    "display_name": "valid metadata2 instance",
    "metadata": {
        "uri_test": "http://localhost:31823/v1/models?version=2017-02-13",
        "number_range_test": 8,
        "required_test": "required_value",
    },
}

another_metadata_json = {
    "schema_name": "metadata-test",
    "name": "another_instance",
    "display_name": "Another Metadata Instance (2)",
    "metadata": {"uri_test": "http://localhost:8081/", "required_test": "required_value"},
}

invalid_metadata_json = {
    "schema_name": "metadata-test",
    "display_name": "Invalid Metadata Instance - bad uri",
    "metadata": {"uri_test": "//localhost:8081/", "required_test": "required_value"},
}

invalid_json = "{\
    'schema_name': 'metadata-test',\
    'display_name': 'Invalid Metadata Instance - missing comma'\
    'metadata': {\
        'uri_test': '//localhost:8081/',\
        'required_test': 'required_value'\
    }\
}"

invalid_no_display_name_json = {
    "schema_name": "metadata-test",
    "metadata": {"uri_test": "//localhost:8081/", "required_test": "required_value"},
}

valid_display_name_json = {
    "schema_name": "metadata-test",
    "display_name": '1 teste "rápido"',
    "metadata": {"required_test": "required_value"},
}


invalid_schema_name_json = {
    "schema_name": "metadata-testxxx",
    "display_name": "invalid schema name",
    "metadata": {
        "uri_test": "http://localhost:31823/v1/models?version=2017-02-13",
        "number_range_test": 8,
        "required_test": "required_value",
    },
}


# Contains all values corresponding to test schema...
complete_metadata_json = {
    "schema_name": "metadata-test",
    "display_name": "complete metadata instance",
    "metadata": {
        "required_test": "required_value",
        "uri_test": "http://localhost:31823/v1/models?version=2017-02-13",
        "integer_exclusivity_test": 7,
        "integer_multipleOf7_test": 42,
        "number_range_test": 8,
        # purposely missing "number_default_test": 42
        "const_test": 3.14,
        "string_length_test": "1234567",
        "enum_test": "rocks",
        "array_test": ["elyra", "rocks", "the", "world"],
        "object_test": {
            "property1": "first prop",
            "property2": "second_prop",
            "property3": "third prop",
            "property4": "fourth prop",
        },
        "boolean_test": True,
        "null_test": "null",
    },
}

# Minimal json to be built upon for each property test.  Only
# required values are specified.
minimal_metadata_json = {
    "schema_name": "metadata-test",
    "display_name": "complete metadata instance",
    "metadata": {"required_test": "required_value"},
}


# Bring-your-own metadata template used to test hierarchical writes.  The
# display_name field will be updated to reflect the location's instance
# in the hierarchy
byo_metadata_json = {
    "schema_name": "metadata-test",
    "display_name": "location",
    "metadata": {"required_test": "required_value"},
}

# Used in test_update_complex to test --file option
one_of_json = {
    "schema_name": "metadata-test",
    "display_name": "oneOf Testing",
    "metadata": {
        "required_test": "required_value",
        "oneOf_test": {"obj2_prop1": 42, "obj2_prop2": 24, "obj_switch": "obj2"},
    },
}

# Used in test_update_complex to test --allOf_test option (i.e., ovp option)
all_of_json = {
    "obj1_prop1": "allOf-test-val1",
    "obj1_prop2": "allOf-test-val2",
    "obj1_switch": "obj1",
    "obj2_prop1": 42,
    "obj2_prop2": 24,
    "obj2_switch": "obj2",
    "obj3_prop1": 42.7,
    "obj3_prop2": True,
    "obj3_switch": "obj3",
}


def create_json_file(location: Any, file_name: str, content: Dict) -> str:
    return create_file(location, file_name, json.dumps(content))


def create_file(location: Any, file_name: str, content: str) -> str:
    try:
        os.makedirs(location)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    resource = os.path.join(location, file_name)
    with open(resource, "w", encoding="utf-8") as f:
        f.write(content)
    return resource


def create_instance(metadata_store: MetadataStore, location: str, name: str, content: Any) -> str:
    resource = name
    if isinstance(metadata_store, FileMetadataStore):
        if isinstance(content, dict):
            create_json_file(location, name + ".json", content)
        else:
            create_file(location, name + ".json", content)
        resource = os.path.join(location, name + ".json")
    elif isinstance(metadata_store, MockMetadataStore):
        instances = metadata_store.instances
        if instances is None:
            setattr(metadata_store, "instances", dict())
            instances = metadata_store.instances
        if not isinstance(content, dict):
            content = {"display_name": name, "reason": f"JSON failed to load for instance '{name}'"}
        instances[name] = content
    return resource


def get_instance(instances, field, value):
    """Given a list of instances (dicts), return the dictionary where field == value."""
    for inst in instances:
        if inst[field] == value:
            return inst
    assert False, f"Value '{value}' for field '{field}' was not found in instances!"


class PropertyTester(object):
    """Helper class used by elyra_md tests to test each of the properties in the test.json schema."""

    name = None  # prefixed with 'test_' is test name, post-fixed with '_test' is schema property name
    negative_res = False  # expected success of first test
    negative_value = None  # value to use in first test (usually negative test)
    negative_stdout = None  # expected string to find in first test's stdout
    negative_stderr = None  # expected string to find in second test's stdout
    positive_res = True  # expected success of second test
    positive_value = None  # value to use in second test (usually successful)

    def __init__(self, name):
        self.name = name
        self.property = name + "_test"

    def run(self, script_runner, mock_data_dir):
        expected_file = os.path.join(mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE, self.name + ".json")
        # Cleanup from any potential previous failures
        if os.path.exists(expected_file):
            os.remove(expected_file)

        # First test
        ret = script_runner.run(
            "elyra-metadata",
            "create",
            METADATA_TEST_SCHEMASPACE,
            "--schema_name=metadata-test",
            "--name=" + self.name,
            "--display_name=" + self.name,
            "--required_test=required_value",
            "--" + self.property + "=" + str(self.negative_value),
        )

        assert ret.success is self.negative_res
        assert self.negative_stdout in ret.stdout
        assert self.negative_stderr in ret.stderr

        # Second test
        ret = script_runner.run(
            "elyra-metadata",
            "create",
            METADATA_TEST_SCHEMASPACE,
            "--schema_name=metadata-test",
            "--name=" + self.name,
            "--display_name=" + self.name,
            "--required_test=required_value",
            "--" + self.property + "=" + str(self.positive_value),
        )

        assert ret.success is self.positive_res
        assert "Metadata instance '" + self.name + "' for schema 'metadata-test' has been written" in ret.stdout

        assert os.path.isdir(os.path.join(mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE))
        assert os.path.isfile(expected_file)

        with open(expected_file, "r") as fd:
            instance_json = json.load(fd)
            assert instance_json["schema_name"] == "metadata-test"
            assert instance_json["display_name"] == self.name
            assert instance_json["metadata"][self.property] == self.positive_value


class MockMetadataStore(MetadataStore):
    """Hypothetical class used to demonstrate (and test) use of custom storage classes."""

    def __init__(self, schemaspace: str, **kwargs: Any) -> None:
        super().__init__(schemaspace, **kwargs)
        self.instances = None

    def schemaspace_exists(self) -> bool:
        """Returns True if the schemaspace for this instance exists"""
        return self.instances is not None

    def fetch_instances(self, name: Optional[str] = None, include_invalid: bool = False) -> List[dict]:
        """Fetch metadata instances"""

        if name:
            if self.instances is not None and name in self.instances:
                instance = self.instances.get(name)
                if instance.get("reason"):
                    raise ValueError(instance.get("reason"))
                instance["name"] = name
                return [instance]
            raise MetadataNotFoundError(self.schemaspace, name)

        # all instances are wanted, filter based on include-invalid and reason ...
        instance_list = []
        if self.instances is not None:
            for name, instance in self.instances.items():
                if include_invalid or not instance.get("reason"):
                    instance["name"] = name
                    instance_list.append(instance)
        return instance_list

    def store_instance(self, name: str, metadata: dict, for_update: bool = False) -> dict:
        """Stores the named metadata instance."""
        try:
            instance = self.fetch_instances(name)
            if not for_update:  # Create - already exists
                raise MetadataExistsError(self.schemaspace, instance[0].get("resource"))
        except MetadataNotFoundError as mnfe:
            if for_update:  # Update - doesn't exist
                raise mnfe from mnfe

        if self.instances is None:
            self.instances = dict()
        self.instances[name] = metadata  # persisted, now fetch

        instance = self.fetch_instances(name)  # confirm persistence
        return instance[0]

    def delete_instance(self, metadata: dict) -> None:
        """Deletes the metadata instance."""
        name = metadata.get("name")
        self.instances.pop(name)


class MockMetadataTest(Metadata):
    """Hypothetical class used to demonstrate (and test) use of custom instance classes.

    This class name is referenced in the metadata-test schema.
    """

    pre_property = None
    post_property = None

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.pre_property = kwargs.get("pre_property")
        self.post_property = kwargs.get("post_property")

    def to_dict(self, trim: bool = False) -> dict:
        d = super().to_dict(trim=trim)
        if self.pre_property is not None:
            d["pre_property"] = self.pre_property
        if self.post_property is not None:
            d["post_property"] = self.post_property
        return d

    def on_load(self, **kwargs: Any) -> None:
        super().on_load(**kwargs)
        self.post_property = self.display_name

    def pre_save(self, **kwargs: Any) -> None:
        super().pre_save(**kwargs)
        self.pre_property = self.metadata["required_test"]

    def post_save(self, **kwargs: Any) -> None:
        super().post_save(**kwargs)
        self.post_property = self.display_name

    def pre_delete(self, **kwargs: Any) -> None:
        super().pre_delete(**kwargs)
        self.pre_property = self.metadata["required_test"]

    def post_delete(self, **kwargs: Any) -> None:
        super().post_delete(**kwargs)
        self.post_property = self.display_name


class MockMetadataTestRollback(MockMetadataTest):
    """Used by metadata tests to validate rollback behaviors when post-save/delete hooks throw exceptions."""

    def post_save(self, **kwargs: Any) -> None:
        super().post_save(**kwargs)
        for_update = kwargs["for_update"]
        if os.getenv("METADATA_TEST_HOOK_OP", "skipped") == "create" and not for_update:
            raise NotImplementedError
        if os.getenv("METADATA_TEST_HOOK_OP", "skipped") == "update" and for_update:
            raise ModuleNotFoundError
        self.post_property = self.display_name

    def post_delete(self, **kwargs: Any) -> None:
        super().post_delete(**kwargs)
        if os.getenv("METADATA_TEST_HOOK_OP", "skipped") == "delete":
            raise FileNotFoundError
        self.post_property = self.display_name


class MockMetadataTestInvalid(object):
    """Invalid metadata instance class that doesn't derive from Metadata.

    This requires an update to schema and is only used for manual testing.
    """

    def __init__(self, **kwargs: Any) -> None:
        pass


class MetadataTestSchemaspace(Schemaspace):
    def __init__(self, *args, **kwargs):
        super().__init__(
            schemaspace_id=METADATA_TEST_SCHEMASPACE_ID,
            name=METADATA_TEST_SCHEMASPACE,
            description="Schemaspace for instances of metadata for testing",
            **kwargs,
        )


class BYOSchemaspaceBadId(Schemaspace):
    def __init__(self, *args, **kwargs):
        super().__init__(schemaspace_id="byo_schemaspace_bad_id", name="byo-schemaspace-bad-id", **kwargs)


class BYOSchemaspaceBadName(Schemaspace):
    def __init__(self, *args, **kwargs):
        super().__init__(
            schemaspace_id="b5b391d7-24f5-4b62-93bb-5e5423e651b8", name="byo.schemaspace-bad.name", **kwargs
        )


class BYOSchemaspaceBadClass(LoggingConfigurable):
    """Class is not a subclass of Schemaspace"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class BYOSchemaspaceCaseSensitiveName(Schemaspace):
    def __init__(self, *args, **kwargs):
        super().__init__(
            schemaspace_id="1b1e461a-c7fa-40f2-a3a3-bf1f2fd48eeA", name="byo-schemaspace_CaseSensitiveName", **kwargs
        )


class BYOSchemaspaceThrows(Schemaspace):
    BYO_SCHEMASPACE_ID = "20c98d38-36f6-4f05-a4dc-9b0a6c2cb734"
    BYO_SCHEMASPACE_NAME = "byo-schemaspace-throws"

    def __init__(self, *args, **kwargs):
        super().__init__(
            schemaspace_id=BYOSchemaspace.BYO_SCHEMASPACE_ID, name=BYOSchemaspace.BYO_SCHEMASPACE_NAME, **kwargs
        )
        raise NotImplementedError("Test that throw from constructor is not harmful.")


class BYOSchemaspace(Schemaspace):
    BYO_SCHEMASPACE_ID = "20c98d38-36f6-4f05-a4dc-9b0a6c2cb733"
    BYO_SCHEMASPACE_NAME = "byo-schemaspace"

    def __init__(self, *args, **kwargs):
        super().__init__(
            schemaspace_id=BYOSchemaspace.BYO_SCHEMASPACE_ID, name=BYOSchemaspace.BYO_SCHEMASPACE_NAME, **kwargs
        )


class MetadataTestSchemasProvider(SchemasProvider):
    """Returns schemas relative to Runtime Images schemaspace."""

    def get_schemas(self) -> List[Dict]:
        schemas = []
        parent_dir = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
        schema_dir = os.path.join(parent_dir, "metadata", "schemas")
        schema_files = [
            json_file
            for json_file in os.listdir(schema_dir)
            if json_file.endswith(".json") and json_file.startswith("metadata-test")
        ]
        for json_file in schema_files:
            schema_file = os.path.join(schema_dir, json_file)
            with io.open(schema_file, "r", encoding="utf-8") as f:
                schema_json = json.load(f)

                if json_file == "metadata-test.json":  # Apply filtering
                    # Update multipleOf from 7 to 6 and and value 'added' to enum-valued property
                    multiple_of: int = schema_json["properties"]["metadata"]["properties"]["integer_multiple_test"][
                        "multipleOf"
                    ]
                    assert multiple_of == 7
                    schema_json["properties"]["metadata"]["properties"]["integer_multiple_test"]["multipleOf"] = 6

                    enum: list = schema_json["properties"]["metadata"]["properties"]["enum_test"]["enum"]
                    assert len(enum) == 2
                    enum.append("added")
                    schema_json["properties"]["metadata"]["properties"]["enum_test"]["enum"] = enum

                schemas.append(schema_json)

        return schemas


def schema_factory(schemaspace_id: str, schemaspace_name: str, num_good: int, bad_reasons: List[str]) -> List[Dict]:
    # get the metadata test schema as a primary copy
    parent_dir = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
    schema_file = os.path.join(parent_dir, "metadata", "schemas", "metadata-test.json")
    with io.open(schema_file, "r", encoding="utf-8") as f:
        primary_schema = json.load(f)

    def create_base_schema(primary: Dict, tag: str, ss_name: str, ss_id: str) -> Dict:
        base_schema: Dict = copy.deepcopy(primary)
        base_schema["title"] = f"BYO Test {tag}"
        base_schema["name"] = f"byo-test-{tag}"
        base_schema["display_name"] = base_schema["title"]
        base_schema["schemaspace"] = ss_name
        base_schema["schemaspace_id"] = ss_id
        base_schema["properties"]["schema_name"]["const"] = base_schema["name"]
        base_schema.pop("metadata_class_name")
        return base_schema

    schemas = []
    # Gather bad schemas
    for reason in bad_reasons:
        schema = create_base_schema(primary_schema, reason, schemaspace_name, schemaspace_id)
        if reason == "missing_required":  # remove display_name
            schema["properties"].pop("display_name")  # This will trigger a validation error
        elif reason == "unknown_schemaspace":  # update schemaspace_id to a non-existent schemaspace
            schema["schemaspace_id"] = NON_EXISTENT_SCHEMASPACE_ID
        schemas.append(schema)

    # Gather good schemas
    for i in range(num_good):
        schemas.append(create_base_schema(primary_schema, str(i), schemaspace_name, schemaspace_id))

    return schemas


class BYOSchemasProvider(SchemasProvider):
    """Test SchemasProvider that loads the metadata-test schema and adjusts its values to match BYOSchemaspace."""

    def get_schemas(self) -> List[Dict]:
        # We'll create 2 good schemas and 2 bad schemas for BYOSchemaspace
        schemas = schema_factory(
            BYOSchemaspace.BYO_SCHEMASPACE_ID,
            BYOSchemaspace.BYO_SCHEMASPACE_NAME,
            2,
            ["missing_required", "unknown_schemaspace"],
        )
        return schemas


class BYOSchemasProviderThrows(SchemasProvider):
    """Test SchemasProvider that raises an exception to ensure the exception doesn't mess things up."""

    def get_schemas(self) -> List[Dict]:
        raise ModuleNotFoundError("Exception to ensure bad providers are not side-effecting.")


class BYOSchemasProviderBadClass(object):
    """Test SchemasProvider that is of the wrong subclass."""

    def get_schemas(self) -> List[Dict]:
        schemas = schema_factory(BYOSchemaspace.BYO_SCHEMASPACE_ID, BYOSchemaspace.BYO_SCHEMASPACE_NAME, 2, [])
        return schemas



#
# Copyright 2018-2023 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import copy
import json
import os
import shutil

from jupyter_server.utils import url_path_join
import pytest
from tornado.httpclient import HTTPClientError

from elyra.metadata.schema import METADATA_TEST_SCHEMASPACE
from elyra.metadata.schema import METADATA_TEST_SCHEMASPACE_ID
from elyra.tests.metadata.test_utils import byo_metadata_json
from elyra.tests.metadata.test_utils import create_json_file
from elyra.tests.metadata.test_utils import get_instance
from elyra.tests.metadata.test_utils import invalid_metadata_json
from elyra.tests.metadata.test_utils import valid_metadata_json
from elyra.tests.util.handlers_utils import expected_http_error

os.environ["METADATA_TESTING"] = "1"  # Enable metadata-tests schemaspace


async def test_bogus_schemaspace(jp_fetch, bogus_location):
    # Validate missing is not found.  Remove the bogus location to ensure its not created
    shutil.rmtree(bogus_location)
    with pytest.raises(HTTPClientError) as e:
        await jp_fetch("elyra", "metadata", "bogus", "missing")

    assert expected_http_error(e, 400)
    assert not os.path.exists(bogus_location)


async def test_missing_instance(jp_fetch, setup_data):
    # Validate missing is not found
    name = "missing"
    with pytest.raises(HTTPClientError) as e:
        await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE, name)
    assert expected_http_error(e, 404)


async def test_invalid_instance(jp_fetch, setup_data):
    # Validate invalid throws 404 with validation message
    with pytest.raises(HTTPClientError) as e:
        await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE, "invalid")
    assert expected_http_error(e, 400)


async def test_valid_instance(jp_fetch, setup_data):
    # Ensure valid metadata can be found
    r = await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE_ID, "valid")

    assert r.code == 200
    metadata = json.loads(r.body.decode())
    assert "schema_name" in metadata
    assert metadata["display_name"] == "valid metadata instance"


async def test_get_instances(jp_fetch, setup_data):
    # Ensure all valid metadata can be found
    r = await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE_ID)
    assert r.code == 200
    metadata = json.loads(r.body.decode())
    assert isinstance(metadata, dict)
    assert len(metadata) == 1
    instances = metadata[METADATA_TEST_SCHEMASPACE_ID]
    assert len(instances) == 2
    assert isinstance(instances, list)
    assert get_instance(instances, "name", "another")
    assert get_instance(instances, "name", "valid")


async def test_get_empty_schemaspace_instances(jp_fetch, schemaspace_location, setup_data):
    # Delete the metadata dir contents and attempt listing metadata
    shutil.rmtree(schemaspace_location)
    r = await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE)
    assert r.code == 200
    metadata = json.loads(r.body.decode())
    assert isinstance(metadata, dict)
    assert len(metadata) == 1
    instances = metadata[METADATA_TEST_SCHEMASPACE]
    assert len(instances) == 0

    # Now create empty schemaspace
    os.makedirs(schemaspace_location)
    r = await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE_ID)
    assert r.code == 200
    metadata = json.loads(r.body.decode())
    assert isinstance(metadata, dict)
    assert len(metadata) == 1
    instances = metadata[METADATA_TEST_SCHEMASPACE_ID]
    assert len(instances) == 0


async def test_get_hierarchy_instances(jp_fetch, setup_hierarchy):
    # Ensure all valid metadata can be found
    r = await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE)
    assert r.code == 200
    metadata = json.loads(r.body.decode())
    assert isinstance(metadata, dict)
    assert len(metadata) == 1
    instances = metadata[METADATA_TEST_SCHEMASPACE]
    assert len(instances) == 3
    assert isinstance(instances, list)
    assert get_instance(instances, "name", "byo_1")
    assert get_instance(instances, "name", "byo_2")
    assert get_instance(instances, "name", "byo_3")
    byo_3 = get_instance(instances, "name", "byo_3")
    assert byo_3["display_name"] == "factory"


async def test_create_instance(jp_base_url, jp_fetch):
    """Create a simple instance - not conflicting with factory instances."""

    valid = copy.deepcopy(valid_metadata_json)
    valid["name"] = "valid"
    body = json.dumps(valid)

    r = await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE_ID, body=body, method="POST")
    assert r.code == 201
    assert r.headers.get("Location") == url_path_join(
        jp_base_url, "/elyra", "metadata", METADATA_TEST_SCHEMASPACE_ID, "valid"
    )
    metadata = json.loads(r.body.decode())
    # Add expected "extra" fields to 'valid' so whole-object comparison is satisfied.
    # These are added during the pre_save(), post_save() and on_load() hooks on the
    # MockMetadataTest class instance or when default values for missing properties are applied.
    valid["pre_property"] = valid["metadata"]["required_test"]
    valid["post_property"] = valid["display_name"]
    valid["metadata"]["number_default_test"] = 42
    assert metadata == valid


async def test_create_hierarchy_instance(jp_fetch, setup_hierarchy):
    """Attempts to create an instance from one in the hierarchy."""

    byo_instance = copy.deepcopy(byo_metadata_json)
    byo_instance["display_name"] = "user"
    byo_instance["name"] = "byo_2"
    body = json.dumps(byo_instance)

    with pytest.raises(HTTPClientError) as e:
        await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE, body=body, method="POST")
    assert expected_http_error(e, 409)

    # Confirm the instance was not changed
    r = await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE)
    assert r.code == 200
    metadata = json.loads(r.body.decode())
    assert isinstance(metadata, dict)
    assert len(metadata) == 1
    instances = metadata[METADATA_TEST_SCHEMASPACE]
    assert len(instances) == 3
    assert isinstance(instances, list)
    byo_2 = get_instance(instances, "name", "byo_2")
    assert byo_2["display_name"] == "factory"


async def test_create_invalid_instance(jp_fetch):
    """Create a simple instance - not conflicting with factory instances."""

    invalid = copy.deepcopy(invalid_metadata_json)
    invalid["name"] = "invalid"
    body = json.dumps(invalid)

    with pytest.raises(HTTPClientError) as e:
        await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE, body=body, method="POST")
    assert expected_http_error(e, 400)


async def test_create_instance_missing_schema(jp_fetch, schemaspace_location):
    """Attempt to create an instance using an invalid schema"""

    missing_schema = copy.deepcopy(valid_metadata_json)
    missing_schema["name"] = "missing_schema"
    missing_schema["schema_name"] = "missing_schema"
    missing_schema.pop("display_name")
    body = json.dumps(missing_schema)

    with pytest.raises(HTTPClientError) as e:
        await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE, body=body, method="POST")
    assert expected_http_error(e, 404)

    # Ensure instance was not created.  Can't use REST here since it will correctly trigger 404
    # even though an instance was created and not removed due to failure to validate (due to
    # missing schema).  Fixed by trapping the FileNotFoundError raised due to no schema.
    assert not os.path.exists(os.path.join(schemaspace_location, "missing_schema.json"))

    with pytest.raises(HTTPClientError) as e:
        await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE, "missing_schema")
    assert expected_http_error(e, 404)


async def test_update_non_existent(jp_fetch, schemaspace_location):
    """Attempt to update a non-existent instance."""

    # Try to update a non-existent instance - 404 expected...
    valid = copy.deepcopy(valid_metadata_json)
    valid["name"] = "valid"
    valid["metadata"]["number_range_test"] = 7
    body = json.dumps(valid)

    # Update (non-existent) instance
    with pytest.raises(HTTPClientError) as e:
        await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE, "valid", body=body, method="PUT")
    assert expected_http_error(e, 404)


async def test_update_instance(jp_fetch, schemaspace_location):
    """Update a simple instance."""

    # Create an instance, then update
    create_json_file(schemaspace_location, "valid.json", valid_metadata_json)
    valid = copy.deepcopy(valid_metadata_json)
    valid["name"] = "valid"
    valid["metadata"]["number_range_test"] = 7
    body = json.dumps(valid)

    # Update instance
    r = await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE_ID, "valid", body=body, method="PUT")
    assert r.code == 200
    instance = json.loads(r.body.decode())
    assert instance["metadata"]["number_range_test"] == 7

    # Confirm update via jp_fetch
    r = await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE, "valid")
    assert r.code == 200
    instance = json.loads(r.body.decode())
    assert instance["metadata"]["number_range_test"] == 7


async def test_invalid_update(jp_fetch, schemaspace_location):
    """Update a simple instance with invalid metadata."""

    # Create an instance, then update with invalid metadata
    create_json_file(schemaspace_location, "update_bad_md.json", valid_metadata_json)

    # Fetch it to get the valid instance
    r = await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE, "update_bad_md")
    assert r.code == 200
    instance = json.loads(r.body.decode())

    # Now attempt the update with bad metadata and ensure previous still exists
    valid2 = copy.deepcopy(valid_metadata_json)
    valid2["name"] = "valid"
    valid2["metadata"]["number_range_test"] = 42
    body2 = json.dumps(valid2)

    with pytest.raises(HTTPClientError) as e:
        await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE_ID, "update_bad_md", body=body2, method="PUT")
    assert expected_http_error(e, 400)

    # Fetch again and ensure it matches the previous instance
    r = await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE_ID, "update_bad_md")
    assert r.code == 200
    instance2 = json.loads(r.body.decode())
    assert instance2 == instance


async def test_update_fields(jp_fetch, schemaspace_location):
    # Create an instance, then update with a new field
    create_json_file(schemaspace_location, "update_fields.json", valid_metadata_json)
    valid = copy.deepcopy(valid_metadata_json)
    valid["metadata"]["number_range_test"] = 7
    body = json.dumps(valid)

    # Update instance adding number_range_test
    r = await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE, "update_fields", body=body, method="PUT")
    assert r.code == 200
    instance = json.loads(r.body.decode())
    assert instance["metadata"]["number_range_test"] == 7

    # Add a new field (per schema) and remove another -
    valid["metadata"].pop("number_range_test")
    valid["metadata"]["string_length_test"] = "valid len"
    body = json.dumps(valid)

    r = await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE, "update_fields", body=body, method="PUT")
    assert r.code == 200
    instance = json.loads(r.body.decode())
    assert instance["metadata"]["string_length_test"] == "valid len"
    assert "number_range_test" not in instance["metadata"]


async def test_update_hierarchy_instance(jp_fetch, setup_hierarchy):
    """Update a simple instance - that's conflicting with factory instances."""

    # Do not name intentionally, since this is an update
    byo_instance = copy.deepcopy(byo_metadata_json)
    byo_instance["display_name"] = "user"
    byo_instance["metadata"]["number_range_test"] = 7
    body = json.dumps(byo_instance)

    # Because this is considered an update, replacement is enabled.
    r = await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE, "byo_2", body=body, method="PUT")
    assert r.code == 200

    # Confirm the instances and ensure byo_2 is in USER area
    r = await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE)
    assert r.code == 200
    metadata = json.loads(r.body.decode())
    assert isinstance(metadata, dict)
    assert len(metadata) == 1
    instances = metadata[METADATA_TEST_SCHEMASPACE]
    assert len(instances) == 3
    assert isinstance(instances, list)
    byo_2 = get_instance(instances, "name", "byo_2")
    assert byo_2["schema_name"] == byo_metadata_json["schema_name"]
    assert byo_2["metadata"]["number_range_test"] == 7

    # Attempt to rename the resource, exception expected.
    byo_2["name"] = "byo_2_renamed"
    body = json.dumps(byo_2)

    with pytest.raises(HTTPClientError) as e:
        await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE, "byo_2", body=body, method="PUT")
    assert expected_http_error(e, 400)

    # Confirm no update occurred
    r = await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE, "byo_2")
    assert r.code == 200
    instance = json.loads(r.body.decode())
    assert instance["name"] == "byo_2"


async def test_delete_instance(jp_fetch, schemaspace_location, setup_data):
    """Create a simple instance - not conflicting with factory instances and delete it."""

    # First, attempt to delete non-existent resource, exception expected.
    with pytest.raises(HTTPClientError) as e:
        await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE, "missing", method="DELETE")
    assert expected_http_error(e, 404)

    create_json_file(schemaspace_location, "valid.json", valid_metadata_json)

    r = await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE, "valid", method="DELETE")
    assert r.code == 204

    # Confirm deletion
    with pytest.raises(HTTPClientError) as e:
        await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE, "valid", method="DELETE")
    assert expected_http_error(e, 404)


async def test_delete_hierarchy_instance(jp_fetch, schemaspace_location, setup_hierarchy):
    """Create a simple instance - that conflicts with factory instances and delete it only if local."""

    with pytest.raises(HTTPClientError) as e:
        await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE, "byo_2", method="DELETE")
    assert expected_http_error(e, 403)

    # create local instance, delete should succeed
    create_json_file(schemaspace_location, "byo_2.json", byo_metadata_json)

    r = await jp_fetch("elyra", "metadata", METADATA_TEST_SCHEMASPACE, "byo_2", method="DELETE")
    assert r.code == 204


async def test_bogus_schema(jp_fetch):
    # Validate missing is not found

    # Remove self.request (and other 'self.' prefixes) once transition to jupyter_server occurs
    with pytest.raises(HTTPClientError) as e:
        await jp_fetch("elyra", "schema", "bogus")
    assert expected_http_error(e, 404)


async def test_missing_runtimes_schema(jp_fetch):
    # Validate missing is not found
    with pytest.raises(HTTPClientError) as e:
        await jp_fetch("elyra", "schema", "runtimes", "missing")
    assert expected_http_error(e, 404)


async def test_get_runtimes_schemas(jp_fetch):
    # Ensure all schema for runtimes can be found
    await _get_schemaspace_schemas(jp_fetch, "runtimes", ["kfp"])


async def test_get_code_snippets_schemas(jp_fetch):
    # Ensure all schema for code-snippets can be found
    await _get_schemaspace_schemas(jp_fetch, "code-snippets", ["code-snippet"])


async def test_get_test_schemas(jp_fetch):
    # Ensure all schema for metadata_tests can be found
    await _get_schemaspace_schemas(jp_fetch, METADATA_TEST_SCHEMASPACE, ["metadata-test", "metadata-test2"])


async def test_get_runtimes_schema(jp_fetch):
    # Ensure all schema for runtimes can be found
    await _get_schemaspace_schema(jp_fetch, "runtimes", "kfp")


async def test_get_code_snippets_schema(jp_fetch):
    # Ensure all schema for code-snippets can be found
    await _get_schemaspace_schema(jp_fetch, "code-snippets", "code-snippet")


async def test_get_test_schema(jp_fetch):
    # Ensure all schema for metadata-test can be found
    await _get_schemaspace_schema(jp_fetch, METADATA_TEST_SCHEMASPACE, "metadata-test")


async def _get_schemaspace_schemas(jp_fetch, schemaspace, expected):
    r = await jp_fetch(
        "elyra",
        "schema",
        schemaspace,
    )
    assert r.code == 200
    schemaspace_schemas = json.loads(r.body.decode())
    assert isinstance(schemaspace_schemas, dict)
    assert len(schemaspace_schemas) == 1
    schemas = schemaspace_schemas[schemaspace]
    assert len(schemas) == len(expected)
    for expected_schema in expected:
        assert get_instance(schemas, "name", expected_schema)


async def _get_schemaspace_schema(jp_fetch, schemaspace, expected):
    r = await jp_fetch("elyra", "schema", schemaspace, expected)
    assert r.code == 200
    schemaspace_schema = json.loads(r.body.decode())
    assert isinstance(schemaspace_schema, dict)
    assert expected == schemaspace_schema["name"]
    assert schemaspace == schemaspace_schema["schemaspace"]


async def test_get_schemaspaces(jp_fetch):
    expected_schemaspaces = ["runtimes", "code-snippets"]

    r = await jp_fetch("elyra", "schemaspace")
    assert r.code == 200
    schemaspaces = json.loads(r.body.decode("utf-8"))
    assert isinstance(schemaspaces, dict)
    assert len(schemaspaces["schemaspaces"]) >= len(expected_schemaspaces)

    for expected_schemaspace in expected_schemaspaces:
        assert expected_schemaspace in schemaspaces["schemaspaces"]


async def test_get_schemaspace_info(jp_fetch):
    r = await jp_fetch("elyra", "schemaspace", METADATA_TEST_SCHEMASPACE)
    assert r.code == 200
    schemaspace_info = json.loads(r.body.decode("utf-8"))
    assert "name" in schemaspace_info
    assert "id" in schemaspace_info
    assert "display_name" in schemaspace_info
    assert "description" in schemaspace_info
    assert "schemas" not in schemaspace_info
    assert schemaspace_info["name"] == METADATA_TEST_SCHEMASPACE
    assert schemaspace_info["id"] == METADATA_TEST_SCHEMASPACE_ID


async def test_get_missing_schemaspace_info(jp_fetch):
    with pytest.raises(HTTPClientError) as e:
        await jp_fetch("elyra", "schemaspace", "missing-schemaspace")
    assert expected_http_error(e, 404)



#
# Copyright 2018-2023 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from pathlib import Path
import tarfile

import pytest

from elyra.pipeline.kfp.processor_kfp import KfpPipelineProcessor
from elyra.pipeline.pipeline import GenericOperation
from elyra.pipeline.properties import ElyraProperty
from elyra.pipeline.registry import PipelineProcessorRegistry


# ---------------------------------------------------
# Fixtures
# ---------------------------------------------------


@pytest.fixture
def processor() -> KfpPipelineProcessor:
    """
    Instantiate a Kubeflow Pipelines processor.
    """
    root_dir = str((Path(__file__).parent / "..").resolve())
    processor = KfpPipelineProcessor(root_dir=root_dir)
    return processor


# ---------------------------------------------------
# Tests for  RuntimePipelineProcessor._get_metadata_configuration
# ---------------------------------------------------


def test_fail_get_metadata_configuration_invalid_namespace(processor: KfpPipelineProcessor):
    with pytest.raises(RuntimeError):
        processor._get_metadata_configuration(schemaspace="non_existent_namespace", name="non_existent_metadata")


# ---------------------------------------------------
# Tests for  RuntimePipelineProcessor._generate_dependency_archive
# ---------------------------------------------------


def test_generate_dependency_archive(processor: KfpPipelineProcessor):
    pipelines_test_file = str((Path(__file__).parent / "resources" / "archive" / "test.ipynb").resolve())
    pipeline_dependencies = ["airflow.json"]
    correct_filelist = ["test.ipynb", "airflow.json"]
    component_properties = {
        "filename": pipelines_test_file,
        "dependencies": pipeline_dependencies,
        "runtime_image": "tensorflow/tensorflow:latest",
    }
    test_operation = GenericOperation(
        id="123e4567-e89b-12d3-a456-426614174000",
        type="execution-node",
        classifier="execute-notebook-node",
        name="test",
        component_props=component_properties,
    )

    archive_location = processor._generate_dependency_archive(test_operation)

    tar_content = []
    with tarfile.open(archive_location, "r:gz") as tar:
        for tarinfo in tar:
            if tarinfo.isreg():
                tar_content.append(tarinfo.name)

    assert sorted(correct_filelist) == sorted(tar_content)


def test_fail_generate_dependency_archive(processor: KfpPipelineProcessor):
    pipelines_test_file = "this/is/a/rel/path/test.ipynb"
    pipeline_dependencies = ["non_existent_file.json"]
    component_properties = {
        "filename": pipelines_test_file,
        "dependencies": pipeline_dependencies,
        "runtime_image": "tensorflow/tensorflow:latest",
    }
    test_operation = GenericOperation(
        id="123e4567-e89b-12d3-a456-426614174000",
        type="execution-node",
        classifier="execute-notebook-node",
        name="test",
        component_props=component_properties,
    )

    with pytest.raises(Exception):
        processor._generate_dependency_archive(test_operation)


# ---------------------------------------------------
# Tests for  RuntimePipelineProcessor._get_dependency_source_dir
# ---------------------------------------------------


def test_get_dependency_source_dir(processor: KfpPipelineProcessor):
    pipelines_test_file = "this/is/a/rel/path/test.ipynb"
    processor.root_dir = "/this/is/an/abs/path/"
    correct_filepath = "/this/is/an/abs/path/this/is/a/rel/path"
    component_properties = {"filename": pipelines_test_file, "runtime_image": "tensorflow/tensorflow:latest"}
    test_operation = GenericOperation(
        id="123e4567-e89b-12d3-a456-426614174000",
        type="execution-node",
        classifier="execute-notebook-node",
        name="test",
        component_props=component_properties,
    )

    filepath = processor._get_dependency_source_dir(test_operation)

    assert filepath == correct_filepath


# ---------------------------------------------------
# Tests for  RuntimePipelineProcessor._get_dependency_archive_name
# ---------------------------------------------------


def test_get_dependency_archive_name(processor: KfpPipelineProcessor):
    pipelines_test_file = "this/is/a/rel/path/test.ipynb"
    correct_filename = "test-this-is-a-test-id.tar.gz"
    component_properties = {"filename": pipelines_test_file, "runtime_image": "tensorflow/tensorflow:latest"}
    test_operation = GenericOperation(
        id="this-is-a-test-id",
        type="execution-node",
        classifier="execute-notebook-node",
        name="test",
        component_props=component_properties,
    )

    filename = processor._get_dependency_archive_name(test_operation)

    assert filename == correct_filename


# ---------------------------------------------------
# Tests for  RuntimePipelineProcessor._collect_envs
# ---------------------------------------------------


def test_collect_envs(processor: KfpPipelineProcessor):
    pipelines_test_file = "this/is/a/rel/path/test.ipynb"

    # add system-owned envs with bogus values to ensure they get set to system-derived values,
    # and include some user-provided edge cases
    operation_envs = [
        {"env_var": "ELYRA_RUNTIME_ENV", "value": '"bogus_runtime"'},
        {"env_var": "ELYRA_ENABLE_PIPELINE_INFO", "value": '"bogus_pipeline"'},
        {"env_var": "ELYRA_WRITABLE_CONTAINER_DIR", "value": ""},  # simulate operation reference in pipeline
        {"env_var": "AWS_ACCESS_KEY_ID", "value": '"bogus_key"'},
        {"env_var": "AWS_SECRET_ACCESS_KEY", "value": '"bogus_secret"'},
        {"env_var": "USER_EMPTY_VALUE", "value": "  "},
        {"env_var": "USER_TWO_EQUALS", "value": "KEY=value"},
        {"env_var": "USER_NO_VALUE", "value": ""},
    ]
    converted_envs = ElyraProperty.create_instance("env_vars", operation_envs)

    test_operation = GenericOperation(
        id="this-is-a-test-id",
        type="execution-node",
        classifier="execute-notebook-node",
        name="test",
        component_props={"filename": pipelines_test_file, "runtime_image": "tensorflow/tensorflow:latest"},
        elyra_props={"env_vars": converted_envs},
    )

    envs = processor._collect_envs(test_operation, cos_secret=None, cos_username="Alice", cos_password="secret")

    assert envs["ELYRA_RUNTIME_ENV"] == "kfp"
    assert envs["AWS_ACCESS_KEY_ID"] == "Alice"
    assert envs["AWS_SECRET_ACCESS_KEY"] == "secret"
    assert envs["ELYRA_ENABLE_PIPELINE_INFO"] == "True"
    assert envs["ELYRA_WRITABLE_CONTAINER_DIR"] == "/tmp"
    assert "USER_EMPTY_VALUE" not in envs
    assert envs["USER_TWO_EQUALS"] == "KEY=value"
    assert "USER_NO_VALUE" not in envs

    # Repeat with non-None secret - ensure user and password envs are not present, but others are
    envs = processor._collect_envs(test_operation, cos_secret="secret", cos_username="Alice", cos_password="secret")

    assert envs["ELYRA_RUNTIME_ENV"] == "kfp"
    assert "AWS_ACCESS_KEY_ID" not in envs
    assert "AWS_SECRET_ACCESS_KEY" not in envs
    assert envs["ELYRA_ENABLE_PIPELINE_INFO"] == "True"
    assert envs["ELYRA_WRITABLE_CONTAINER_DIR"] == "/tmp"
    assert "USER_EMPTY_VALUE" not in envs
    assert envs["USER_TWO_EQUALS"] == "KEY=value"
    assert "USER_NO_VALUE" not in envs


# ---------------------------------------------------
# Tests for  RuntimePipelineProcessor._process_list_value_function
# ---------------------------------------------------


def test_process_list_value_function(processor: KfpPipelineProcessor):
    # Test values that will be successfully converted to list
    assert processor._process_list_value("") == []
    assert processor._process_list_value(None) == []
    assert processor._process_list_value("[]") == []
    assert processor._process_list_value("None") == []
    assert processor._process_list_value("['elem1']") == ["elem1"]
    assert processor._process_list_value("['elem1', 'elem2', 'elem3']") == ["elem1", "elem2", "elem3"]
    assert processor._process_list_value("  ['elem1',   'elem2' , 'elem3']  ") == ["elem1", "elem2", "elem3"]
    assert processor._process_list_value("[1, 2]") == [1, 2]
    assert processor._process_list_value("[True, False, True]") == [True, False, True]
    assert processor._process_list_value("[{'obj': 'val', 'obj2': 'val2'}, {}]") == [{"obj": "val", "obj2": "val2"}, {}]

    # Test values that will not be successfully converted to list
    assert processor._process_list_value("[[]") == "[[]"
    assert processor._process_list_value("[elem1, elem2]") == "[elem1, elem2]"
    assert processor._process_list_value("elem1, elem2") == "elem1, elem2"
    assert processor._process_list_value("  elem1, elem2  ") == "elem1, elem2"
    assert processor._process_list_value("'elem1', 'elem2'") == "'elem1', 'elem2'"


# ---------------------------------------------------
# Tests for  RuntimePipelineProcessor._process_dictionary_value
# ---------------------------------------------------


def test_process_dictionary_value_function(processor: KfpPipelineProcessor):
    # Test values that will be successfully converted to dictionary
    assert processor._process_dictionary_value("") == {}
    assert processor._process_dictionary_value(None) == {}
    assert processor._process_dictionary_value("{}") == {}
    assert processor._process_dictionary_value("None") == {}
    assert processor._process_dictionary_value("{'key': 'value'}") == {"key": "value"}

    dict_as_str = "{'key1': 'value', 'key2': 'value'}"
    assert processor._process_dictionary_value(dict_as_str) == {"key1": "value", "key2": "value"}

    dict_as_str = "  {  'key1': 'value'  , 'key2'  : 'value'}  "
    assert processor._process_dictionary_value(dict_as_str) == {"key1": "value", "key2": "value"}

    dict_as_str = "{'key1': [1, 2, 3], 'key2': ['elem1', 'elem2']}"
    assert processor._process_dictionary_value(dict_as_str) == {"key1": [1, 2, 3], "key2": ["elem1", "elem2"]}

    dict_as_str = "{'key1': 2, 'key2': 'value', 'key3': True, 'key4': None, 'key5': [1, 2, 3]}"
    expected_value = {"key1": 2, "key2": "value", "key3": True, "key4": None, "key5": [1, 2, 3]}
    assert processor._process_dictionary_value(dict_as_str) == expected_value

    dict_as_str = "{'key1': {'key2': 2, 'key3': 3, 'key4': 4}, 'key5': {}}"
    expected_value = {
        "key1": {
            "key2": 2,
            "key3": 3,
            "key4": 4,
        },
        "key5": {},
    }
    assert processor._process_dictionary_value(dict_as_str) == expected_value

    # Test values that will not be successfully converted to dictionary
    assert processor._process_dictionary_value("{{}") == "{{}"
    assert processor._process_dictionary_value("{key1: value, key2: value}") == "{key1: value, key2: value}"
    assert processor._process_dictionary_value("  { key1: value, key2: value }  ") == "{ key1: value, key2: value }"
    assert processor._process_dictionary_value("key1: value, key2: value") == "key1: value, key2: value"
    assert processor._process_dictionary_value("{'key1': true}") == "{'key1': true}"
    assert processor._process_dictionary_value("{'key': null}") == "{'key': null}"

    dict_as_str = "{'key1': [elem1, elem2, elem3], 'key2': ['elem1', 'elem2']}"
    assert processor._process_dictionary_value(dict_as_str) == dict_as_str

    dict_as_str = "{'key1': {key2: 2}, 'key3': ['elem1', 'elem2']}"
    assert processor._process_dictionary_value(dict_as_str) == dict_as_str

    dict_as_str = "{'key1': {key2: 2}, 'key3': ['elem1', 'elem2']}"
    assert processor._process_dictionary_value(dict_as_str) == dict_as_str


@pytest.fixture
def processor_registry(monkeypatch, expected_runtimes) -> PipelineProcessorRegistry:
    PipelineProcessorRegistry.clear_instance()
    if expected_runtimes:
        monkeypatch.setenv("ELYRA_PROCESSOR_RUNTIMES", ",".join(expected_runtimes))
    ppr = PipelineProcessorRegistry.instance()
    yield ppr
    PipelineProcessorRegistry.clear_instance()


@pytest.mark.parametrize(
    "expected_runtimes",
    [
        ["local"],
        ["kfp"],
        ["local"],
        None,
    ],
)
def test_processor_registry_filtering(expected_runtimes, processor_registry):
    if expected_runtimes is None:
        expected_runtimes = ["local", "kfp"]
    runtimes = processor_registry.get_all_processors()
    assert len(runtimes) == len(expected_runtimes)
    for runtime in runtimes:
        assert runtime.name in expected_runtimes



#
# Copyright 2018-2023 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
from pathlib import Path

import pytest
from requests import session

from elyra.util.url import FileTransportAdapter
from elyra.util.url import get_verify_parm


def test_valid_file_url():
    """
    Verify that the FileTransportAdapter works as expected for valid input
    """
    requests_session = session()
    requests_session.mount("file://", FileTransportAdapter())

    # utilize the test source code as resource
    this_file = __file__
    url = Path(this_file).as_uri()
    res = requests_session.get(url)
    assert res.status_code == 200, this_file
    with open(this_file, "r") as source:
        assert res.text == source.read()


def test_invalid_file_url():
    """
    Verify that the FileTransportAdapter works as expected for error
    scenarios:
     - requested resource is a directory
     - requested resource not found
     - request method is not 'GET'
    """
    requests_session = session()
    requests_session.mount("file://", FileTransportAdapter())

    # utilize the test source code as resource
    this_file_p = Path(__file__)

    # requested resource is a directory
    url = this_file_p.parent.as_uri()
    res = requests_session.get(url)
    assert res.status_code == 400, url
    assert res.reason == "Not a file"
    assert len(res.text) == 0

    # requested resource wasn't found
    url = this_file_p.resolve().with_name("no-such-file").as_uri()
    res = requests_session.get(url)
    assert res.status_code == 404, url
    assert res.reason == "File not found"
    assert len(res.text) == 0

    # request method is not supported
    url = this_file_p.as_uri()
    unsupported_methods = [
        requests_session.delete,
        requests_session.head,
        requests_session.options,
        requests_session.patch,
        requests_session.post,
        requests_session.put,
    ]
    for unsupported_method in unsupported_methods:
        res = unsupported_method(url)
        assert res.status_code == 405, url
        assert res.reason == "Method not allowed"


@pytest.fixture
def setup_env_vars():
    # runs before test (save the value of environment variable
    # TRUSTED_CA_BUNDLE_PATH to avoid any contamination by tests
    # that modify it)
    current_TRUSTED_CA_BUNDLE_PATH_value = os.environ.get("TRUSTED_CA_BUNDLE_PATH")
    yield
    # runs after test (restore the value of environment variable
    # TRUSTED_CA_BUNDLE_PATH, if it was defined)
    if current_TRUSTED_CA_BUNDLE_PATH_value is not None:
        os.environ["TRUSTED_CA_BUNDLE_PATH"] = current_TRUSTED_CA_BUNDLE_PATH_value


@pytest.mark.usefixtures("setup_env_vars")
def test_valid_get_verify_parm():
    """
    Verify that method get_verify_parm works as expected:
     - env variable TRUSTED_CA_BUNDLE_PATH is defined
     - env variable TRUSTED_CA_BUNDLE_PATH is not defined, but a default is specified
     - env variable TRUSTED_CA_BUNDLE_PATH is not defined and no default is specified
    """
    test_TRUSTED_CA_BUNDLE_PATH_value = "/path/to/cert/bundle"
    os.environ["TRUSTED_CA_BUNDLE_PATH"] = test_TRUSTED_CA_BUNDLE_PATH_value
    assert get_verify_parm() == test_TRUSTED_CA_BUNDLE_PATH_value
    del os.environ["TRUSTED_CA_BUNDLE_PATH"]
    # set explicit default
    assert get_verify_parm(False) is False
    # set explicit default
    assert get_verify_parm(True) is True
    # use implicit default
    assert get_verify_parm() is True


@pytest.mark.usefixtures("setup_env_vars")
def test_invalid_get_verify_parm():
    """
    Verify that method get_verify_parm works as if environment variable
    TRUSTED_CA_BUNDLE_PATH contains an invalid value
    """
    test_TRUSTED_CA_BUNDLE_PATH_value = ""
    os.environ["TRUSTED_CA_BUNDLE_PATH"] = test_TRUSTED_CA_BUNDLE_PATH_value
    assert get_verify_parm() is True

    test_TRUSTED_CA_BUNDLE_PATH_value = "   "
    os.environ["TRUSTED_CA_BUNDLE_PATH"] = test_TRUSTED_CA_BUNDLE_PATH_value
    assert get_verify_parm() is True
